# Copyright (C) 2021, Camilo Hincapié Gutiérrez
# This file is part of CDSLIB.
#
# CDSLIB is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CDSLIB is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
#
# This package is authored by:
# Camilo Hincapié (https://www.linkedin.com/in/camilo-hincapie-gutierrez/) (main author)
# Ian Mejía (https://github.com/IanMejia)
# Emil Rueda (https://www.linkedin.com/in/emil-rueda-424012207/)
# Nicole Rivera (https://github.com/nicolerivera1)
# Carolina Rojas Duque (https://github.com/carolinarojasd)

from datetime import timedelta
from typing import Union, Optional
from copy import deepcopy

from numpy import where, full, isin, concatenate, setdiff1d, array
from numpy import isnan, nan, transpose, equal
from numpy.random import choice, random_sample
from pandas.core.frame import DataFrame
from pandas.core.series import Series
from pandas import concat
from dask.dataframe import from_pandas

from abmodel.utils import ExecutionModes
from abmodel.utils import check_field_existance
from abmodel.utils import exception_burner
from abmodel.utils import std_str_join_cols
from abmodel.models import DistTitles
from abmodel.models import NaturalHistory
from abmodel.models import DiseaseStates
from abmodel.models import SusceptibilityGroups
from abmodel.models import ImmunizationGroups
from abmodel.models import MobilityGroups
from abmodel.models import IsolationAdherenceGroups
from abmodel.models import MRAdherenceGroups
from abmodel.models import HealthSystem
from abmodel.models import InterestVariables
from abmodel.models import MRTStopModes
from abmodel.models import CyclicMRModes
from abmodel.models import GlobalCyclicMR


# =============================================================================
def init_calculate_max_time_iterative(
    key: str,
    natural_history: NaturalHistory
) -> bool:
    """
        TODO: Add brief explanation

        Parameters
        ----------
        TODO

        Returns
        -------
        TODO

        Notes
        -----
        TODO: include mathematical description and explanatory image

        See Also
        --------
        init_calculate_max_time_vectorized : TODO complete explanation

        Examples
        --------
        TODO: include some examples
    """
    dist_type = natural_history.items[key] \
        .dist[DistTitles.time.value].dist_type
    return False if dist_type is None else True


def init_calculate_max_time_vectorized(
    key: Series,  # str
    natural_history: NaturalHistory
) -> Series:
    """
        TODO: Add brief explanation

        Parameters
        ----------
        TODO

        Returns
        -------
        TODO

        Notes
        -----
        TODO: include mathematical description and explanatory image

        See Also
        --------
        init_calculate_max_time_iterative : TODO complete explanation

        Examples
        --------
        TODO: include some examples
    """
    return list(map(
        lambda single_key: False if natural_history
        .items[single_key]
        .dist[DistTitles.time.value].dist_type is None else True,
        key
    ))


# =============================================================================
def calculate_max_time_iterative(
    key: str,
    disease_state: str,
    do_calculate_max_time: bool,
    disease_state_time: Union[float, None],
    disease_state_max_time: Union[float, None],
    disease_groups: DiseaseStates,
    natural_history: NaturalHistory
) -> Series([float, float]):
    """
        TODO: Add brief explanation

        Parameters
        ----------
        TODO

        Returns
        -------
        TODO

        Notes
        -----
        TODO: include mathematical description and explanatory image

        See Also
        --------
        calculate_max_time_vectorized : TODO complete explanation

        Examples
        --------
        TODO: include some examples
    """
    if do_calculate_max_time:
        if disease_groups.items[disease_state].is_dead:
            return Series([nan, nan])
        else:
            return Series([0,
                           natural_history.items[key]
                           .dist[DistTitles.time.value]
                           .sample()
                           ])
    else:
        return Series([disease_state_time, disease_state_max_time])


# TODO: Reimplement this function in order to also return disease_state_time
def calculate_max_time_vectorized(
    key: Series,  # str
    do_calculate_max_time: Series,  # bool
    disease_state_time: Series,  # float or None
    disease_state_max_time: Series,  # float or None
    natural_history: NaturalHistory
) -> Series:  # Series of float
    """
        TODO: Add brief explanation

        Parameters
        ----------
        TODO

        Returns
        -------
        TODO

        Notes
        -----
        TODO: include mathematical description and explanatory image

        See Also
        --------
        calculate_max_time_iterative : TODO complete explanation

        Examples
        --------
        TODO: include some examples
    """
    return list(map(
        lambda single_key, cond, time_value: natural_history
        .items[single_key]
        .dist[DistTitles.time.value]
        .sample() if cond else time_value,
        zip(key, do_calculate_max_time, disease_state_max_time)
    ))


# =============================================================================
def transition_function(
    disease_state: str,
    disease_state_time: float,
    disease_state_max_time: float,
    is_dead: bool,
    key: str,
    disease_groups: DiseaseStates,
    natural_history: NaturalHistory
) -> Series([str, float, bool, bool, bool]):
    """
        TODO: Add brief explanation

        Parameters
        ----------
        TODO

        Returns
        -------
        TODO

        Notes
        -----
        TODO: include mathematical description and explanatory image

        Examples
        --------
        TODO: include some examples
    """
    # Set False the flag that forces to calculate
    # disease state max time and the flag that forces to update immunization
    # params. They are going to be changed to True if disease state changes
    do_calculate_max_time = False
    do_update_immunization_params = False

    if not isnan(disease_state_max_time):
        if disease_state_time >= disease_state_max_time:
            # Get the transitions info for the current
            # key (vulnerability_group, current disease_state)
            transitions = natural_history.items[key].transitions

            # Get disease_states enabled for the probable transition and
            # their corresponding probabilities
            disease_states = list(transitions.keys())
            probabilities = [
                transitions[transition].probability
                for transition in disease_states
                ]

            # disease state must change
            # Verify: becomes into? ... Throw the dice
            disease_state = choice(
                disease_states,
                p=probabilities
                )

            # Set True the flag that forces to calculate
            # disease state max time
            do_calculate_max_time = True

            # Set True the flag that forces to update
            # immunization params
            do_update_immunization_params = True

            # Update is_dead value
            is_dead = disease_groups.items[disease_state].is_dead

    return Series([disease_state, disease_state_time, is_dead,
                   do_calculate_max_time, do_update_immunization_params])


# =============================================================================
def hospitalization_vectorized(
    is_hospitalized: Series,
    is_in_ICU: Series,
    disease_states: Series,
    is_dead: Series,
    reduction_factor: Series,
    dead_disease_group: str,
    alpha: float,  # Reduction factor of spread prob due to hospitalization
    disease_groups: DiseaseStates,
    health_system: HealthSystem
) -> DataFrame:
    """
        TODO: Add brief explanation
        # step 1: no is_hospitalized and disease_state has probability
        # to be hospitalized

        # step 2: is_hospitalized ... No matter if disease states
        # changes or not ... Throw dice to see if agent still
        # hospitalized

        # step 3: if one agent cannot be hospitalized, see if it dies
        # because of disease

        Parameters
        ----------
        TODO

        Returns
        -------
        TODO

        Notes
        -----
        TODO: include mathematical description and explanatory image

        Examples
        --------
        TODO: include some examples
    """
    # Get former is_hospitalized info
    agents_number = len(is_hospitalized)

    # Please be aware that if an agent is in ICU, then it must be
    # hospitalized too
    former_is_hospitalized = deepcopy(is_hospitalized)
    former_is_in_ICU = deepcopy(is_in_ICU)

    # ICU state has higher priority
    # Verify: is going to be in ICU or remains in ICU?
    # ... Throw the dice ... Do it for all the agents
    dice = random_sample(agents_number)

    ICU_prob = array([
        disease_groups
        .items[disease_state]
        .dist[DistTitles.icu_prob.value]
        .sample()
        for disease_state in disease_states
        ])
    ICU_prob[equal(ICU_prob, None)] = 0

    # if dice <= ICU_prob
    # agent should be in ICU
    is_in_ICU = where(
        dice <= ICU_prob,
        full(agents_number, True),
        full(agents_number, False)
        )

    new_in_ICU_number = len(
        where(equal(is_in_ICU, True))[0]
        )

    if new_in_ICU_number > health_system.ICU_capacity:
        # Health system got overloaded in ICU
        # Some agents are susceptible to die
        # 1. Those agents that were in ICU before, should remain in ICU
        # 2. New agents in ICU, must throw the dice to see who dies
        # or who survives.

        # NOTE: if on the contrary, health system still having
        # ICU capacity, then this ICU vacancy can be used for
        # hospitalization if the health system gets overloaded
        # for hospitalization, in other words, hospitalization
        # vacancy includes ICU vacancy, but the contrary doesn't hold.
        # This is simulated enabling that those in ICU are also
        # hospitalized

        completely_new_in_ICU = where(
            (equal(is_in_ICU, True)) & (equal(former_is_in_ICU, False))
            )[0]

        must_die_number = new_in_ICU_number \
            - health_system.ICU_capacity

        # Update new_in_ICU_number
        new_in_ICU_number = health_system.ICU_capacity

        must_die = choice(
            completely_new_in_ICU,
            size=must_die_number,
            replace=False
            )

        # Update disease state of those that died
        disease_states[must_die] = dead_disease_group

        # Update is_dead for those that died
        is_dead[must_die] = True

        is_in_ICU_False = where(equal(is_in_ICU, False))[0]
        is_in_ICU_True = where(equal(is_in_ICU, True))[0]

        new_is_in_ICU_False = concatenate([is_in_ICU_False, must_die])
        mask = isin(is_in_ICU_True, must_die, invert=True)
        new_is_in_ICU_True = is_in_ICU_True[mask]

        # Update values
        is_in_ICU[new_is_in_ICU_False] = False
        is_in_ICU[new_is_in_ICU_True] = True

    # Next step is to change hospitalization state
    # Verify: is going to be hospitalized or remains hospitalized?
    # ... Throw the dice ... Do it for all the agents
    dice = random_sample(agents_number)

    hospitalization_prob = array([
        disease_groups
        .items[disease_state]
        .dist[DistTitles.hospitalization.value]
        .sample()
        for disease_state in disease_states
        ])
    hospitalization_prob[equal(hospitalization_prob, None)] = 0

    # if dice <= hospitalization_prob
    # agent should be hospitalized
    is_hospitalized = where(
        dice <= hospitalization_prob,
        full(agents_number, True),
        full(agents_number, False)
        )

    # Merge this new is_hospitalized with those
    # new hospitalized due to ICU
    # Remember that those in ICU are also hospitalized
    is_hospitalized = where(
        (equal(is_hospitalized, False)) & (equal(is_in_ICU, True)),
        full(agents_number, True),
        is_hospitalized
        )

    new_hospitalized_number = len(
        where(equal(is_hospitalized, True))[0]
        )

    if new_hospitalized_number > health_system.hospital_capacity:
        # Health system got overloaded in hospital vacancy
        # Some agents are susceptible to die
        # 1. ICU has higher priority. Those agents that are in ICU,
        # should remain in ICU
        # 2. New hospitalized agents, that are not in ICU, must throw
        # the dice to see who dies or who survives.
        completely_new_hospitalized_not_in_ICU = where(
            (equal(is_hospitalized, True)) & (equal(is_in_ICU, False))
            & (equal(former_is_hospitalized, False))
            )[0]

        susceptible_to_die_number = len(
            completely_new_hospitalized_not_in_ICU
            )

        must_die_number = new_hospitalized_number \
            - health_system.hospital_capacity

        if must_die_number < susceptible_to_die_number:
            must_die = choice(
                completely_new_hospitalized_not_in_ICU,
                size=must_die_number,
                replace=False
                )
        elif susceptible_to_die_number == must_die_number:
            must_die = completely_new_hospitalized_not_in_ICU
        else:
            # must_die_number > susceptible_to_die_number
            # The surplus of must die number forces us to include the
            # possibility that some former hospitalized not in UCI have
            # died
            must_die_number_surplus = must_die_number \
                - susceptible_to_die_number

            remain_hospitalized_but_not_in_ICU = where(
                (equal(is_hospitalized, True)) & (equal(is_in_ICU, False))
                & (equal(former_is_hospitalized, True))
                )[0]

            must_die_surplus = choice(
                remain_hospitalized_but_not_in_ICU,
                size=must_die_number_surplus,
                replace=False
                )

            must_die = concatenate([
                must_die_surplus,
                completely_new_hospitalized_not_in_ICU
            ])

        # Update disease state of those that died
        disease_states[must_die] = dead_disease_group

        # Update is_dead for those that died
        is_dead[must_die] = True

        is_hospitalized_False = where(equal(is_hospitalized, False))[0]
        is_hospitalized_True = where(equal(is_hospitalized, True))[0]

        new_is_hospitalized_False = concatenate([
            is_hospitalized_False, must_die
            ])
        mask = isin(is_hospitalized_True, must_die, invert=True)
        new_is_hospitalized_True = is_hospitalized_True[mask]

        # Update values
        is_hospitalized[new_is_hospitalized_False] = False
        is_hospitalized[new_is_hospitalized_True] = True

    # Determine indexes for changing reduction factor
    new_hospitalized_indexes = where(
        (equal(is_hospitalized, True)) & (equal(former_is_hospitalized, False))
        )[0]

    recovered_from_hospitalization_indexes = where(
        (equal(is_hospitalized, False)) & (equal(former_is_hospitalized, True))
        )[0]

    # Calculate reduction factor
    reduction_factor_copy = reduction_factor.copy()

    reduction_factor_copy[new_hospitalized_indexes] = \
        reduction_factor[new_hospitalized_indexes]*alpha

    reduction_factor_copy[recovered_from_hospitalization_indexes] = \
        reduction_factor[recovered_from_hospitalization_indexes]/alpha

    # Wrap data together
    data = array([is_hospitalized, is_in_ICU, disease_states, is_dead,
                 reduction_factor_copy])
    data_transposed = transpose(data)

    return DataFrame(data_transposed)


# =============================================================================
def diagnosis_function(
    disease_state: str,
    is_dead: bool,
    is_diagnosed: bool,
    disease_groups: DiseaseStates
) -> bool:
    """
        TODO: Add brief explanation

        Parameters
        ----------
        TODO

        Returns
        -------
        TODO

        Notes
        -----
        TODO: include mathematical description and explanatory image

        Examples
        --------
        TODO: include some examples
    """
    if is_dead:
        is_diagnosed = False
    else:
        if is_diagnosed:
            # is_diagnosed = True
            # do nothing
            pass
        else:
            # it is not diagnosed
            is_infected = disease_groups \
                .items[disease_state].is_infected
            if is_infected:
                # Agent can be diagnosed
                # Verify: is going to be diagnosed? ... Throw the dice
                dice = random_sample()

                be_diagnosed_prob = disease_groups \
                    .items[disease_state] \
                    .dist[DistTitles.diagnosis.value] \
                    .sample()

                if dice <= be_diagnosed_prob:
                    # Agent was diagnosed !!!
                    is_diagnosed = True
                else:
                    # Agent was not diagnosed
                    # do nothing
                    pass
            else:
                is_diagnosed = False
    return is_diagnosed


# =============================================================================
def isolation_function(
    disease_state: str,
    disease_groups: DiseaseStates
) -> tuple[bool, float, float]:
    """
        TODO: Add brief explanation

        Parameters
        ----------
        TODO

        Returns
        -------
        TODO

        Notes
        -----
        TODO: include mathematical description and explanatory image

        Examples
        --------
        TODO: include some examples
    """
    # How much time is going to be isolated?
    # ... Throw the dice
    # isolation_max_time is in the scale of days
    isolation_max_time = disease_groups \
        .items[disease_state] \
        .dist[DistTitles.isolation_days.value] \
        .sample()

    isolation_time = 0.0
    is_isolated = True

    return is_isolated, isolation_time, isolation_max_time


def isolation_handler(
    disease_state: str,
    isolation_adherence_group: str,
    is_diagnosed: bool,
    is_isolated: bool,
    isolation_time: float,
    isolation_max_time: float,
    adheres_to_isolation: bool,
    reduction_factor: float,
    beta: float,  # Reduction factor of spread prob due to being isolated
    disease_groups: DiseaseStates,
    isolation_adherence_groups: Optional[IsolationAdherenceGroups] = None
) -> Series([bool, bool, float, float, bool, float]):
    """
        TODO: Add brief explanation

        Parameters
        ----------
        TODO

        Returns
        -------
        TODO

        Notes
        -----
        TODO: include mathematical description and explanatory image

        See Also
        --------
        isolation_function : TODO complete explanation

        Examples
        --------
        TODO: include some examples
    """
    if not is_diagnosed:
        # is_diagnosed = False
        # do nothing
        pass
    else:
        # it is diagnosed
        if is_isolated:
            # it is isolated
            if isolation_time >= isolation_max_time:
                # isolation must end
                isolation_time = nan
                isolation_max_time = nan
                is_isolated = False
                is_diagnosed = False
                reduction_factor = reduction_factor/beta
            else:
                # isolation has not finished yet
                # do nothing
                pass
        else:
            # it is not isolated
            if isolation_adherence_groups is None:
                # Agent always adheres to isolation
                adheres_to_isolation = True
                reduction_factor = reduction_factor*beta

                (is_isolated, isolation_time,
                    isolation_max_time) = isolation_function(
                    disease_state,
                    disease_groups
                )
            else:
                # isolation_adherence_groups is not None

                # Does agent adhere to be isolated?
                # ... Throw the dice
                dice = random_sample()

                adherence_prob = isolation_adherence_groups.items[
                    isolation_adherence_group
                    ].dist[
                        DistTitles.adherence.value
                        ].sample()

                if dice <= adherence_prob:
                    # Agent adheres to isolation
                    adheres_to_isolation = True
                    reduction_factor = reduction_factor*beta

                    (is_isolated, isolation_time,
                        isolation_max_time) = isolation_function(
                        disease_state,
                        disease_groups
                    )
                else:
                    # Agent doesn't adhere to isolation
                    # do nothing
                    adheres_to_isolation = False

    return Series([is_diagnosed, is_isolated, isolation_time,
                   isolation_max_time, adheres_to_isolation,
                   reduction_factor])


def mr_handler(
    mr_group: str,
    mr_adherence_group: str,
    is_diagnosed: bool,
    reduction_factor: float,
    beta: float,  # Reduction factor of spread prob due to being isolated
    mrc_target_groups: list,
    mr_adherence_groups: Optional[IsolationAdherenceGroups] = None
) -> Series([bool, bool, float]):
    """
        TODO: Add brief explanation

        Parameters
        ----------
        TODO

        Returns
        -------
        TODO

        Notes
        -----
        TODO: include mathematical description and explanatory image

        See Also
        --------
        isolation_function : TODO complete explanation

        Examples
        --------
        TODO: include some examples
    """
    if mr_group not in mrc_target_groups:
        # it is not isolated by mr
        adheres_to_mr_isolation = False
        isolated_by_mr = False
    else:
        # it is isolated by mr
        isolated_by_mr = True
        if mr_adherence_groups is None:
            # Agent always adheres to isolation
            adheres_to_mr_isolation = True
            reduction_factor = reduction_factor*beta
        else:
            # mr_adherence_groups is not None

            # Does agent adhere to be isolated?
            # ... Throw the dice
            dice = random_sample()

            adherence_prob = mr_adherence_groups.items[
                mr_adherence_group
                ].dist[
                    DistTitles.mr_adherence.value
                    ].sample()

            if dice <= adherence_prob:
                # Agent adheres to isolation
                adheres_to_mr_isolation = True
                if is_diagnosed:
                    reduction_factor = reduction_factor*beta
            else:
                # Agent doesn't adhere to isolation
                adheres_to_mr_isolation = False

    return Series([isolated_by_mr, adheres_to_mr_isolation, reduction_factor])


# =============================================================================
def contagion_function(
    agent: int,
    x: float,
    y: float,
    immunization_level: float,
    key: str,
    disease_state: str,
    susceptibility_group: str,
    times_infected: int,
    disease_state_time: float,
    reduction_factor: float,
    natural_history: NaturalHistory,
    disease_groups: DiseaseStates,
    susceptibility_groups: SusceptibilityGroups,
    kdtree_by_disease_state: dict,
    agents_labels_by_disease_state: dict,
    df_copy: DataFrame
) -> Series([str, int, list, float, bool, bool]):
    """
        TODO
        # Set None the disease_state_time
        # It should be changed to zero 0 if gets infected

        TODO: Add brief explanation

        Parameters
        ----------
        TODO

        Returns
        -------
        TODO

        Notes
        -----
        TODO: include mathematical description and explanatory image

        Examples
        --------
        TODO: include some examples
    """
    # Set False the flag that forces to calculate
    # disease state max time and the flag that forces to update immunization
    # params. They are going to be changed to True if disease state changes
    do_calculate_max_time = False
    do_update_immunization_params = False

    # List to save who infected the agent
    infected_by = []

    # Get transition_by_contagion boolean
    transition_by_contagion = \
        natural_history.items[key].transition_by_contagion

    if transition_by_contagion:
        # Get the transitions info for the current
        # key (vulnerability_group, current disease_state)
        transitions = natural_history.items[key].transitions

        # Get disease_states enabled for the probable transition
        # after contagion and and their corresponding probabilities
        disease_states = list(transitions.keys())
        probabilities = [
            transitions[transition].probability
            for transition in disease_states
            ]

        # Agent location
        agent_location = [x, y]

        # Cycle through each spreader to see if the agent gets
        # infected by the spreader
        spreaders = []
        for disease_state_label in disease_groups.items.keys():
            if disease_groups.items[disease_state_label].can_spread:
                spreaders.append(disease_state_label)

        for spreader_state in spreaders:
            if kdtree_by_disease_state[spreader_state]:
                # Retrieve spread_radius
                spread_radius = \
                    disease_groups.items[spreader_state] \
                    .spread_radius

                # Detect if the agents of "disease_state" that are
                # inside a distance equal to the tracing_radius
                # points_inside_radius_array is a ndarray with a
                # list of indeces which correspond neighbors of
                # agent
                points_inside_radius_array = \
                    kdtree_by_disease_state[spreader_state] \
                    .query_ball_point(
                        agent_location,
                        spread_radius
                        )

                # Now we have to get the corresponding spreader
                # labels excluding the agent's own label
                spreader_labels_inside_radius = list(setdiff1d(
                        agents_labels_by_disease_state[
                            spreader_state][
                            points_inside_radius_array],
                        agent
                        )
                    )

                # Calculate joint probability for contagion
                spread_probability = disease_groups.items[spreader_state] \
                    .spread_probability

                joint_probability = \
                    (1.0 - immunization_level) \
                    * (susceptibility_groups.items[susceptibility_group]
                       .dist[DistTitles.susceptibility.value].sample()) \
                    * spread_probability * reduction_factor

                # Check if got infected
                for spreader in spreader_labels_inside_radius:
                    # Throw the dice
                    dice = random_sample()

                    if dice <= joint_probability:
                        # Got infected !!!
                        # Save who infected the agent
                        infected_by.append(spreader)

        if len(infected_by) != 0:
            # Update times_infected
            times_infected += 1

            # Verify: becomes into? ... Throw the dice
            disease_state = choice(
                disease_states,
                p=probabilities
                )

            # Set the time elapsed since it got infected
            # in zero
            disease_state_time = 0

            # Set True the flag that forces to calculate
            # disease state max time
            do_calculate_max_time = True

            # Set True the flag that forces to update
            # immunization params
            do_update_immunization_params = True
    else:
        # transition_by_contagion == False
        # It is supposed that this is the case for those with is_dead == True
        pass

    return Series([disease_state, times_infected, infected_by,
                   disease_state_time, do_calculate_max_time,
                   do_update_immunization_params])


# =============================================================================
def init_immunization_params_iterative(
    immunization_group: str,
    immunization_level: float,
    immunization_groups: ImmunizationGroups
) -> Series([float, float, float]):
    """
        TODO: Add brief explanation

        Parameters
        ----------
        TODO

        Returns
        -------
        TODO

        Notes
        -----
        TODO: include mathematical description and explanatory image

        Examples
        --------
        TODO: include some examples
    """
    if immunization_level != 0:

        distribution = immunization_groups \
            .items[immunization_group] \
            .dist[DistTitles.immunization_time.value]

        if distribution.dist_type is not None:
            # Init immunization_time
            # In scale of days
            immunization_time = 0

            # Calculate immunization_max_time
            # In scale of days
            immunization_max_time = distribution.sample()

            # Calculate immunization_slope
            immunization_slope = - immunization_level/immunization_max_time
        else:
            raise ValueError(
                f"Error with {DistTitles.immunization_time.value} for"
                f"immunization group '{immunization_group}': "
                "If immunization_level != 0"
                f"then {DistTitles.immunization_time.value} must not be None"
                )
    else:
        immunization_time = nan
        immunization_max_time = nan
        immunization_slope = nan

    return Series([immunization_time, immunization_max_time,
                   immunization_slope])


def update_immunization_params_iterative(
    key: str,  # Must correspond to the outdated key
    disease_state: str,
    immunization_level: float,
    immunization_slope: float,
    immunization_time: float,  # In scale of days
    immunization_max_time: float,  # In scale of days
    do_update_immunization_params: bool,
    natural_history: NaturalHistory
) -> Series([float, float, float, float, bool]):
    """
        TODO: Add brief explanation

        Parameters
        ----------
        TODO

        Returns
        -------
        TODO

        Notes
        -----
        TODO: include mathematical description and explanatory image

        Examples
        --------
        TODO: include some examples
    """
    if do_update_immunization_params:
        # Get the transitions info for the former
        # key (vulnerability_group, former disease_state)
        transitions = natural_history.items[key].transitions

        # Update immunization_level with the gain
        gain = transitions[disease_state] \
            .immunization_gain

        immunization_level = min([immunization_level + gain, 1.0])

        # Update immunization_max_time and immunization_slope
        addend = transitions[disease_state] \
            .dist[DistTitles.immunization_time.value] \
            .sample()

        if addend is not None:
            if isnan(immunization_max_time):
                immunization_time_remainder = 0.0
            else:
                immunization_time_remainder = immunization_max_time \
                                            - immunization_time

            immunization_max_time = addend + immunization_time_remainder
        else:
            pass

        if not isnan(immunization_max_time):
            immunization_slope = - immunization_level/immunization_max_time
        else:
            immunization_slope = nan

        # Restart immnuzation_time
        immunization_time = 0

        do_update_immunization_params = False

    return Series([immunization_level, immunization_slope, immunization_time,
                   immunization_max_time, do_update_immunization_params])


# =============================================================================
def update_immunization_level_iterative(
    dt: float,  # In scale of days
    immunization_level: float,
    immunization_slope: float,
    immunization_time: float,  # In scale of days
    immunization_max_time: float  # In scale of days
) -> Series([float, float, float, float]):
    """
        TODO: Add brief explanation

        Parameters
        ----------
        TODO

        Returns
        -------
        TODO

        Notes
        -----
        TODO: include mathematical description and explanatory image

        See Also
        --------
        immunization_function : TODO complete explanation

        Examples
        --------
        TODO: include some examples
    """
    if immunization_time < immunization_max_time:
        immunization_level = immunization_slope * dt + immunization_level
        immunization_time = immunization_time + dt
    else:
        # immunization_time >= immunization_max_time
        immunization_level = 0
        immunization_slope = nan
        immunization_time = nan
        immunization_max_time = nan

    return Series([immunization_level, immunization_slope, immunization_time,
                   immunization_max_time])


# =============================================================================
def alertness_function(
    agent: int,
    key: str,
    x: float,
    y: float,
    is_dead: bool,
    vulnerability_group: str,
    disease_state: str,
    natural_history: NaturalHistory,
    disease_groups: DiseaseStates,
    kdtree_by_disease_state: dict,
    agents_labels_by_disease_state: dict,
    dead_disease_group: str
) -> Series([bool, list]):
    """
        TODO: Add brief explanation

        Parameters
        ----------
        TODO

        Returns
        -------
        TODO

        Notes
        -----
        TODO: include mathematical description and explanatory image

        Examples
        --------
        TODO: include some examples
    """
    is_alert = False
    alerted_by = []

    if is_dead:
        pass
    else:
        all_avoidable_neighbors = array([])

        # Agent location
        agent_location = [x, y]

        # =============================================
        # Cycle through each state of the neighbors to see if the agent
        # should be alert

        for avoidable_state in disease_groups.items.keys():
            if avoidable_state != dead_disease_group:
                # Compute avoidable_agent_key
                avoidable_agent_key = std_str_join_cols(
                    str(vulnerability_group),
                    str(avoidable_state)
                    )

                # Get radius to avoid an avoidable_agent (avoidance_radius)
                avoidance_radius = natural_history \
                    .items[avoidable_agent_key].avoidance_radius

                if (avoidance_radius != 0
                   and kdtree_by_disease_state[avoidable_state]):
                    # Detect if any avoidable agent is inside a distance
                    # equal to the corresponding avoidance_radius
                    points_inside_radius_array = \
                        kdtree_by_disease_state[avoidable_state] \
                        .query_ball_point(
                            agent_location,
                            avoidance_radius
                            )

                    avoidable_neighbors = setdiff1d(
                        agents_labels_by_disease_state[
                            avoidable_state][
                            points_inside_radius_array],
                        agent
                        )

                    if len(avoidable_neighbors) != 0:

                        all_avoidable_neighbors = concatenate(
                            (all_avoidable_neighbors, avoidable_neighbors),
                            axis=None
                            )

                        for avoidable_agent_index in avoidable_neighbors:
                            # Calculate alertness probability
                            probability = natural_history.items[key] \
                                .dist[DistTitles.alertness.value].sample()

                            if probability:

                                # Must agent be alert ? ... Throw the dice
                                dice = random_sample()

                                # Note that alertness depends on a probability,
                                # which tries to model the probability that an
                                # agent with a defined group and state is alert
                                if dice <= probability:

                                    # Agent is alerted !!!
                                    is_alert = True

                                    # Append avoidable_agent_index in
                                    # alerted_by
                                    alerted_by.append(avoidable_agent_index)

        # float to int
        all_avoidable_neighbors = all_avoidable_neighbors.astype(int)

        # ndarray to list
        all_avoidable_neighbors = all_avoidable_neighbors.tolist()

    return Series([is_alert, alerted_by])


# =============================================================================
class AgentDisease:
    """
        TODO: Add brief explanation

        Methods
        -------
        TODO
    """
    @classmethod
    def init_required_fields(
        cls,
        df: DataFrame,
        dead_disease_group: str,
        alpha: float,
        beta: float,
        disease_groups: DiseaseStates,
        natural_history: NaturalHistory,
        health_system: HealthSystem,
        immunization_groups: Optional[ImmunizationGroups] = None,
        isolation_adherence_groups: Optional[IsolationAdherenceGroups] = None,
        execmode: ExecutionModes = ExecutionModes.iterative.value
    ) -> DataFrame:
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            Returns
            -------
            TODO

            Raises
            ------
            TODO

            Notes
            -----
            TODO: include mathematical description and explanatory image

            See Also
            --------
            TODO

            Examples
            --------
            TODO: include some examples
        """
        # Initialize is_dead
        df = cls.init_is_dead(df, disease_groups, execmode)

        # Generate key column
        df = cls.generate_key_col(df, execmode)

        # Init disease_state_max_time
        df = cls.init_disease_state_max_time(df, disease_groups,
                                             natural_history, execmode)

        # Init is_hospìtalized and is_in_ICU
        df = df.assign(is_hospitalized=False)
        df = df.assign(is_in_ICU=False)
        df = df.assign(reduction_factor=1.0)
        df = cls.to_hospitalize_agents(
            df, dead_disease_group, alpha, disease_groups, health_system,
            ExecutionModes.vectorized.value
            )

        # Init is_diagnosed
        df = df.assign(is_diagnosed=False)
        df = cls.to_diagnose_agents(df, disease_groups, execmode)

        # Init is_isolated
        df = df.assign(is_isolated=False)
        df = df.assign(isolation_time=nan)
        df = df.assign(isolation_max_time=nan)
        df = df.assign(adheres_to_isolation=True)
        dt = 0.0  # Set dt to zero for initialization purposes
        df = cls.to_isolate_agents(
            df, dt, beta, disease_groups, isolation_adherence_groups, execmode
            )

        # Init has_mr_restictions
        df = df.assign(isolated_by_mr=False)
        df = df.assign(adheres_to_mr_isolation=True)
        # df = cls.mr_isolate_agents(
        #    df, dt, beta, disease_groups, isolation_adherence_groups, execmode
        #    )

        # Init times_infected
        df = cls.init_times_infected(df, disease_groups, execmode)

        if immunization_groups is not None:
            # Init immunization_level
            df = cls.init_immunization_level(df, immunization_groups, execmode)

            # Init immunization params
            df = cls.init_immunization_params(df, immunization_groups)
        else:
            # Init immunization_level
            df = df.assign(immunization_level=0)

            # Init immunization params
            df = df.assign(immunization_time=nan)
            df = df.assign(immunization_max_time=nan)
            df = df.assign(immunization_slope=nan)

        return df

    @classmethod
    def init_is_dead(
        cls,
        df: DataFrame,
        disease_groups: DiseaseStates,
        execmode: ExecutionModes = ExecutionModes.iterative.value
    ) -> DataFrame:
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            Returns
            -------
            TODO

            Raises
            ------
            TODO

            Notes
            -----
            TODO: include mathematical description and explanatory image

            See Also
            --------
            abmodel.agent.execution_modes.ExecutionModes : TODO complete
            explanation

            check_field_existance : TODO complete explanation

            Examples
            --------
            TODO: include some examples
        """
        try:
            if execmode == ExecutionModes.iterative.value:
                df["is_dead"] = df.apply(
                    lambda row: disease_groups
                    .items[row["disease_state"]].is_dead,
                    axis=1
                )
            elif execmode == ExecutionModes.dask.value:
                df = from_pandas(df, npartitions=1)
                df["is_dead"] = df.apply(
                    lambda row: disease_groups
                    .items[row["disease_state"]].is_dead,
                    axis=1,
                    meta=('is_dead', 'bool')
                )
                df = df.compute()
            else:
                raise NotImplementedError(
                    f"`execmode = {execmode}` is still not implemented yet"
                    )
        except Exception as error:
            validation_list = ["disease_state"]
            exception_burner([
                error,
                check_field_existance(df, validation_list)
                ])
        else:
            return df

    @classmethod
    def init_times_infected(
        cls,
        df: DataFrame,
        disease_groups: DiseaseStates,
        execmode: ExecutionModes = ExecutionModes.iterative.value,
        npartitions: Optional[int] = 1
    ) -> DataFrame:
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            Returns
            -------
            TODO

            Raises
            ------
            TODO

            Notes
            -----
            TODO: include mathematical description and explanatory image

            See Also
            --------
            abmodel.agent.execution_modes.ExecutionModes : TODO complete
            explanation

            check_field_existance : TODO complete explanation

            Examples
            --------
            TODO: include some examples
        """
        try:
            if execmode == ExecutionModes.iterative.value:
                df["times_infected"] = df.apply(
                    lambda row: 1 if disease_groups
                    .items[row["disease_state"]].is_infected else 0,
                    axis=1
                    )
            elif execmode == ExecutionModes.dask.value:
                df = from_pandas(df, npartitions=npartitions)
                df["times_infected"] = df.apply(
                    lambda row: 1 if disease_groups
                    .items[row["disease_state"]].is_infected else 0,
                    axis=1,
                    meta=(0, "int64")
                    )
                df = df.compute()
            else:
                raise NotImplementedError(
                    f"`execmode = {execmode}` is still not implemented yet"
                    )
        except Exception as error:
            validation_list = ["disease_state"]
            exception_burner([
                error,
                check_field_existance(df, validation_list)
                ])
        else:
            return df

    @classmethod
    def init_immunization_level(
        cls,
        df: DataFrame,
        immunization_groups: ImmunizationGroups,
        execmode: ExecutionModes = ExecutionModes.iterative.value,
        npartitions: Optional[int] = 1
    ) -> DataFrame:
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            Returns
            -------
            TODO

            Raises
            ------
            TODO

            Notes
            -----
            TODO: include mathematical description and explanatory image

            See Also
            --------
            abmodel.agent.execution_modes.ExecutionModes : TODO complete
            explanation

            check_field_existance : TODO complete explanation

            Examples
            --------
            TODO: include some examples
        """
        try:
            if execmode == ExecutionModes.iterative.value:
                df["immunization_level"] = df.apply(
                    lambda row: immunization_groups
                    .items[row["immunization_group"]].dist[
                        DistTitles.immunization_level.value].sample(),
                    axis=1
                    )
            elif execmode == ExecutionModes.dask.value:
                df = from_pandas(df, npartitions=npartitions)
                df["immunization_level"] = df.apply(
                    lambda row: immunization_groups
                    .items[row["immunization_group"]].dist[
                        DistTitles.immunization_level.value].sample(),
                    axis=1,
                    meta=(0, "int64")
                    )
                df = df.compute()
            else:
                raise NotImplementedError(
                    f"`execmode = {execmode}` is still not implemented yet"
                    )
        except Exception as error:
            validation_list = ["immunization_group"]
            exception_burner([
                error,
                check_field_existance(df, validation_list)
                ])
        else:
            return df

    @classmethod
    def init_immunization_params(
        cls,
        df: DataFrame,
        immunization_groups: Optional[ImmunizationGroups],
        execmode: ExecutionModes = ExecutionModes.iterative.value,
        npartitions: Optional[int] = 1
    ) -> DataFrame:
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            Returns
            -------
            TODO

            Raises
            ------
            TODO

            Notes
            -----
            TODO: include mathematical description and explanatory image

            See Also
            --------
            abmodel.agent.execution_modes.ExecutionModes : TODO complete
            explanation

            check_field_existance : TODO complete explanation

            Examples
            --------
            TODO: include some examples
        """
        try:
            if execmode == ExecutionModes.iterative.value:
                df[["immunization_time", "immunization_max_time",
                    "immunization_slope"]] = df.apply(
                    lambda row: init_immunization_params_iterative(
                        row["immunization_group"],
                        row["immunization_level"],
                        immunization_groups
                        ),
                    axis=1
                    )
            elif execmode == ExecutionModes.dask.value:
                df = from_pandas(df, npartitions=npartitions)
                df[["immunization_time", "immunization_max_time",
                    "immunization_slope"]] = df.apply(
                    lambda row: init_immunization_params_iterative(
                        row["immunization_group"],
                        row["immunization_level"],
                        immunization_groups
                        ),
                    axis=1,
                    meta={
                        0: "float64",
                        1: "float64",
                        2: "float64"
                    }
                    )
                df = df.compute()
            else:
                raise NotImplementedError(
                    f"`execmode = {execmode}` is still not implemented yet"
                    )
        except Exception as error:
            validation_list = ["immunization_group", "immunization_level"]
            exception_burner([
                error,
                check_field_existance(df, validation_list)
                ])
        else:
            return df

    @classmethod
    def generate_key_col(
        cls,
        df: DataFrame,
        execmode: ExecutionModes = ExecutionModes.iterative.value,
        npartitions: Optional[int] = 1
    ) -> DataFrame:
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            Returns
            -------
            TODO

            Raises
            ------
            TODO

            Notes
            -----
            TODO: include mathematical description and explanatory image

            See Also
            --------
            abmodel.agent.execution_modes.ExecutionModes : TODO complete
            explanation

            std_str_join_cols : TODO complete explanation

            check_field_existance : TODO complete explanation

            Examples
            --------
            TODO: include some examples
        """
        try:
            if execmode == ExecutionModes.iterative.value:
                df["key"] = df.apply(
                    lambda row: std_str_join_cols(
                        str(row["vulnerability_group"]),
                        str(row["disease_state"])
                        ),
                    axis=1
                    )
            elif execmode == ExecutionModes.dask.value:
                df = from_pandas(df, npartitions=npartitions)
                df["key"] = df.apply(
                    lambda row: std_str_join_cols(
                        str(row["vulnerability_group"]),
                        str(row["disease_state"])
                        ),
                    axis=1,
                    meta=(0, "str")
                    )
                df = df.compute()
            elif execmode == ExecutionModes.vectorized.value:
                df["key"] = std_str_join_cols(
                    df["vulnerability_group"],
                    df["disease_state"]
                    )
            else:
                raise NotImplementedError(
                    f"`execmode = {execmode}` is still not implemented yet"
                    )
        except Exception as error:
            validation_list = ["disease_state", "vulnerability_group"]
            exception_burner([
                error,
                check_field_existance(df, validation_list)
                ])
        else:
            return df

    @classmethod
    def init_disease_state_max_time(
        cls,
        df: DataFrame,
        disease_groups: DiseaseStates,
        natural_history: NaturalHistory,
        execmode: ExecutionModes = ExecutionModes.iterative.value,
        npartitions: Optional[int] = 1
    ) -> DataFrame:
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            Returns
            -------
            TODO

            Raises
            ------
            TODO

            Notes
            -----
            TODO: include mathematical description and explanatory image

            See Also
            --------
            abmodel.agent.execution_modes.ExecutionModes : TODO complete
            explanation

            check_field_existance : TODO complete explanation

            init_calculate_max_time_iterative : TODO complete explanation

            init_calculate_max_time_vectorized : TODO complete explanation

            Examples
            --------
            TODO: include some examples
        """
        try:
            # Init disease_state_time and disease_state_max_time with None
            df = df.assign(disease_state_time=nan)
            df = df.assign(disease_state_max_time=nan)

            if execmode == ExecutionModes.iterative.value:
                df["do_calculate_max_time"] = df.apply(
                    lambda row: init_calculate_max_time_iterative(
                        row["key"],
                        natural_history
                        ),
                    axis=1
                    )
            elif execmode == ExecutionModes.vectorized.value:
                df["do_calculate_max_time"] = \
                    init_calculate_max_time_vectorized(
                        df["key"],
                        natural_history
                        )
            else:
                raise NotImplementedError(
                    f"`execmode = {execmode}` is still not implemented yet"
                    )
        except Exception as error:
            validation_list = ["key"]
            exception_burner([
                error,
                check_field_existance(df, validation_list)
                ])
        else:
            # Call determine_disease_state_max_time function in order
            # to calculate it for those agents who are infected
            df = cls.determine_disease_state_max_time(
                df, disease_groups, natural_history, execmode
            )
            return df

    # TODO: Reimplement calculate_max_time_vectorized
    @classmethod
    def determine_disease_state_max_time(
        cls,
        df: DataFrame,
        disease_groups: DiseaseStates,
        natural_history: NaturalHistory,
        execmode: ExecutionModes = ExecutionModes.iterative.value,
        npartitions: Optional[float] = 1
    ) -> DataFrame:
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            Returns
            -------
            TODO

            Raises
            ------
            TODO

            Notes
            -----
            TODO: include mathematical description and explanatory image

            See Also
            --------
            abmodel.agent.execution_modes.ExecutionModes : TODO complete
            explanation

            check_field_existance : TODO complete explanation

            calculate_max_time_iterative : TODO complete explanation

            calculate_max_time_vectorized : TODO complete explanation

            Examples
            --------
            TODO: include some examples
        """
        try:
            if execmode == ExecutionModes.iterative.value:
                df[["disease_state_time",
                    "disease_state_max_time"]] = df.apply(
                    lambda row: calculate_max_time_iterative(
                        row["key"],
                        row["disease_state"],
                        row["do_calculate_max_time"],
                        row["disease_state_time"],
                        row["disease_state_max_time"],
                        disease_groups,
                        natural_history
                        ),
                    axis=1
                    )
            elif execmode == ExecutionModes.dask.value:
                df = from_pandas(df, npartitions=npartitions)
                df[["disease_state_time",
                    "disease_state_max_time"]] = df.apply(
                    lambda row: calculate_max_time_iterative(
                        row["key"],
                        row["disease_state"],
                        row["do_calculate_max_time"],
                        row["disease_state_time"],
                        row["disease_state_max_time"],
                        disease_groups,
                        natural_history
                        ),
                    axis=1,
                    meta={
                        0: "float64",
                        1: "float64"
                    }
                    )
                df = df.compute()
            # TODO: Reimplement calculate_max_time_vectorized
            # elif execmode == ExecutionModes.vectorized.value:
            #     df["disease_state_max_time"] = calculate_max_time_vectorized(
            #         df["key"],
            #         df["do_calculate_max_time"],
            #         df["disease_state_max_time"],
            #         natural_history
            #         )
            else:
                raise NotImplementedError(
                    f"`execmode = {execmode}` is still not implemented yet"
                    )
        except Exception as error:
            validation_list = ["disease_state_max_time", "disease_state_time",
                               "key", "do_calculate_max_time"]
            exception_burner([
                error,
                check_field_existance(df, validation_list)
                ])
        else:
            # Restart do_calculate_max_time
            df = df.assign(do_calculate_max_time=False)

            return df

    @classmethod
    def disease_state_transition(
        cls,
        df: DataFrame,
        dt: float,  # In scale of days
        disease_groups: DiseaseStates,
        natural_history: NaturalHistory,
        execmode: ExecutionModes = ExecutionModes.iterative.value,
        npartitions: Optional[int] = 1
    ) -> DataFrame:
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            Returns
            -------
            TODO

            Raises
            ------
            TODO

            Notes
            -----
            TODO: include mathematical description and explanatory image

            See Also
            --------
            abmodel.agent.execution_modes.ExecutionModes : TODO complete
            explanation

            check_field_existance : TODO complete explanation

            transition_function : TODO complete explanation

            determine_disease_state_max_time : TODO complete explanation

            Examples
            --------
            TODO: include some examples
        """
        try:
            if execmode == ExecutionModes.iterative.value:
                # Update disease state time
                df["disease_state_time"] = df["disease_state_time"] + dt

                df[["disease_state", "disease_state_time", "is_dead",
                    "do_calculate_max_time",
                    "do_update_immunization_params"]] = df.apply(
                    lambda row: transition_function(
                        row["disease_state"],
                        row["disease_state_time"],
                        row["disease_state_max_time"],
                        row["is_dead"],
                        row["key"],
                        disease_groups,
                        natural_history
                        ),
                    axis=1
                    )
            elif execmode == ExecutionModes.dask.value:
                # Update disease state time
                df = from_pandas(df, npartitions=npartitions)
                df["disease_state_time"] = df["disease_state_time"] + dt

                df[["disease_state", "disease_state_time", "is_dead",
                    "do_calculate_max_time",
                    "do_update_immunization_params"]] = df.apply(
                    lambda row: transition_function(
                        row["disease_state"],
                        row["disease_state_time"],
                        row["disease_state_max_time"],
                        row["is_dead"],
                        row["key"],
                        disease_groups,
                        natural_history
                        ),
                    axis=1,
                    meta={
                        0: "str",
                        1: "float64",
                        2: "bool",
                        3: "bool",
                        4: "bool"
                    }
                    )
                df = df.compute()

            else:
                raise NotImplementedError(
                    f"`execmode = {execmode}` is still not implemented yet"
                    )

            # Call determine_disease_state_max_time function in order
            # to calculate it for those agents who went through a transition
            df = cls.determine_disease_state_max_time(
                df, disease_groups, natural_history, execmode
                )

            # Call update_immunization_params function in order to update them
            # for those agents who went through a transition
            df = cls.update_immunization_params(
                df, natural_history, execmode
                )

            # Update key column
            df = cls.generate_key_col(df, execmode)

        except Exception as error:
            validation_list = ["disease_state", "disease_state_time",
                               "is_dead", "key", "disease_state_max_time"]
            exception_burner([
                error,
                check_field_existance(df, validation_list)
                ])
        else:
            return df

    @classmethod
    def to_hospitalize_agents(
        cls,
        df: DataFrame,
        dead_disease_group: str,
        alpha: float,  # Reduction factor of spread prob due to hospitalization
        disease_groups: DiseaseStates,
        health_system: HealthSystem,
        execmode: ExecutionModes = ExecutionModes.vectorized.value
    ) -> DataFrame:
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            Returns
            -------
            TODO

            Raises
            ------
            TODO

            Notes
            -----
            TODO: include mathematical description and explanatory image

            See Also
            --------
            abmodel.agent.execution_modes.ExecutionModes : TODO complete
            explanation

            check_field_existance : TODO complete explanation

            hospitalization_vectorized : TODO complete explanation

            Examples
            --------
            TODO: include some examples
        """
        try:
            if execmode == ExecutionModes.vectorized.value:
                df[["is_hospitalized", "is_in_ICU",
                    "disease_state", "is_dead",
                    "reduction_factor"]] = hospitalization_vectorized(
                    df["is_hospitalized"],
                    df["is_in_ICU"],
                    df["disease_state"],
                    df["is_dead"],
                    df["reduction_factor"],
                    dead_disease_group,
                    alpha,
                    disease_groups,
                    health_system
                    )
            else:
                raise NotImplementedError(
                    f"`execmode = {execmode}` is still not implemented yet"
                    )
        except Exception as error:
            validation_list = ["is_hospitalized", "is_in_ICU", "disease_state",
                               "is_dead", "reduction_factor"]
            exception_burner([
                error,
                check_field_existance(df, validation_list)
                ])
        else:
            return df

    @classmethod
    def to_diagnose_agents(
        cls,
        df: DataFrame,
        disease_groups: DiseaseStates,
        execmode: ExecutionModes = ExecutionModes.iterative.value,
        npartitions: Optional[int] = 1
    ) -> DataFrame:
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            Returns
            -------
            TODO

            Raises
            ------
            TODO

            Notes
            -----
            TODO: include mathematical description and explanatory image

            See Also
            --------
            abmodel.agent.execution_modes.ExecutionModes : TODO complete
            explanation

            check_field_existance : TODO complete explanation

            diagnosis_function : TODO complete explanation

            Examples
            --------
            TODO: include some examples
        """
        try:
            if execmode == ExecutionModes.iterative.value:
                df["is_diagnosed"] = df.apply(
                    lambda row: diagnosis_function(
                        row["disease_state"],
                        row["is_dead"],
                        row["is_diagnosed"],
                        disease_groups
                        ),
                    axis=1
                    )
            elif execmode == ExecutionModes.dask.value:
                df = from_pandas(df, npartitions=npartitions)
                df["is_diagnosed"] = df.apply(
                    lambda row: diagnosis_function(
                        row["disease_state"],
                        row["is_dead"],
                        row["is_diagnosed"],
                        disease_groups
                        ),
                    axis=1,
                    meta=(0, "bool")
                    )
                df = df.compute()
            else:
                raise NotImplementedError(
                    f"`execmode = {execmode}` is still not implemented yet"
                    )
        except Exception as error:
            validation_list = ["disease_state", "is_dead", "is_diagnosed"]
            exception_burner([
                error,
                check_field_existance(df, validation_list)
                ])
        else:
            return df

    @classmethod
    def to_isolate_agents(
        cls,
        df: DataFrame,
        dt: float,  # dt represents the iteration_time in the scale of days,
        beta: float,  # Reduction factor of spread prob due to being isolated
        disease_groups: DiseaseStates,
        isolation_adherence_groups: Optional[IsolationAdherenceGroups] = None,
        execmode: ExecutionModes = ExecutionModes.iterative.value,
        npartitions: Optional[int] = 1
    ) -> DataFrame:
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            Returns
            -------
            TODO

            Raises
            ------
            TODO

            Notes
            -----
            TODO: include mathematical description and explanatory image

            See Also
            --------
            abmodel.agent.execution_modes.ExecutionModes : TODO complete
            explanation

            check_field_existance : TODO complete explanation

            isolation_handler : TODO complete explanation

            Examples
            --------
            TODO: include some examples
        """
        try:
            if execmode == ExecutionModes.iterative.value:
                # Update isolation time
                df["isolation_time"] = df["isolation_time"] + dt

                df[["is_diagnosed", "is_isolated", "isolation_time",
                    "isolation_max_time", "adheres_to_isolation",
                    "reduction_factor"]] = df.apply(
                    lambda row: isolation_handler(
                        row["disease_state"],
                        row["isolation_adherence_group"],
                        row["is_diagnosed"],
                        row["is_isolated"],
                        row["isolation_time"],
                        row["isolation_max_time"],
                        row["adheres_to_isolation"],
                        row["reduction_factor"],
                        beta,
                        disease_groups,
                        isolation_adherence_groups
                        ),
                    axis=1
                    )
            elif execmode == ExecutionModes.dask.value:
                # Update isolation time
                df = from_pandas(df, npartitions=npartitions)
                df["isolation_time"] = df["isolation_time"] + dt

                df[["is_diagnosed", "is_isolated", "isolation_time",
                    "isolation_max_time", "adheres_to_isolation",
                    "reduction_factor"]] = df.apply(
                    lambda row: isolation_handler(
                        row["disease_state"],
                        row["isolation_adherence_group"],
                        row["is_diagnosed"],
                        row["is_isolated"],
                        row["isolation_time"],
                        row["isolation_max_time"],
                        row["adheres_to_isolation"],
                        row["reduction_factor"],
                        beta,
                        disease_groups,
                        isolation_adherence_groups
                        ),
                    axis=1,
                    meta={
                        0: "bool",
                        1: "bool",
                        2: "float64",
                        3: "float64",
                        4: "bool",
                        5: "float64",
                    }
                    )
                df = df.compute()
            else:
                raise NotImplementedError(
                    f"`execmode = {execmode}` is still not implemented yet"
                    )
        except Exception as error:
            validation_list = ["disease_state", "isolation_adherence_group",
                               "is_isolated", "is_diagnosed", "isolation_time",
                               "isolation_max_time", "adheres_to_isolation",
                               "reduction_factor"]
            exception_burner([
                error,
                check_field_existance(df, validation_list)
                ])
        else:
            return df

    @classmethod
    def to_isolate_agents_by_mr(
        cls,
        df: DataFrame,
        mrc_target_groups: list,
        beta: float,  # Reduction factor of spread prob due to being isolated
        mr_adherence_groups: Optional[IsolationAdherenceGroups] = None,
        execmode: ExecutionModes = ExecutionModes.iterative.value,
        npartitions: Optional[int] = 1
    ) -> DataFrame:
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            Returns
            -------
            TODO

            Raises
            ------
            TODO

            Notes
            -----
            TODO: include mathematical description and explanatory image

            See Also
            --------
            abmodel.agent.execution_modes.ExecutionModes : TODO complete
            explanation

            check_field_existance : TODO complete explanation

            isolation_handler : TODO complete explanation

            Examples
            --------
            TODO: include some examples
        """
        try:
            if execmode == ExecutionModes.iterative.value:
                df[["isolated_by_mr", "adheres_to_mr_isolation",
                    "reduction_factor"]] = df.apply(
                    lambda row: mr_handler(
                        row["mr_group"],
                        row["mr_adherence_group"],
                        row["is_diagnosed"],
                        row["reduction_factor"],
                        beta,
                        mrc_target_groups,
                        mr_adherence_groups
                        ),
                    axis=1
                    )
            elif execmode == ExecutionModes.dask.value:
                df = from_pandas(df, npartitions=npartitions)
                df[["isolated_by_mr", "adheres_to_mr_isolation",
                    "reduction_factor"]] = df.apply(
                    lambda row: mr_handler(
                        row["mr_group"],
                        row["mr_adherence_group"],
                        row["is_diagnosed"],
                        row["reduction_factor"],
                        beta,
                        mrc_target_groups,
                        mr_adherence_groups
                        ),
                    axis=1,
                    meta={
                        0: "bool",
                        1: "bool",
                        2: "float64"
                    }
                    )
                df = df.compute()
            else:
                raise NotImplementedError(
                    f"`execmode = {execmode}` is still not implemented yet"
                    )
        except Exception as error:
            validation_list = ["mr_group", "mr_adherence_group",
                               "is_diagnosed", "isolated_by_mr",
                               "adheres_to_mr_isolation", "reduction_factor"]
            exception_burner([
                error,
                check_field_existance(df, validation_list)
                ])
        else:
            return df

    @classmethod
    def disease_state_transition_by_contagion(
        cls,
        df: DataFrame,
        kdtree_by_disease_state: dict,
        agents_labels_by_disease_state: dict,
        natural_history: NaturalHistory,
        disease_groups: DiseaseStates,
        susceptibility_groups: SusceptibilityGroups,
        execmode: ExecutionModes = ExecutionModes.iterative.value,
        npartitions: Optional[int] = 1
    ) -> DataFrame:
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            Returns
            -------
            TODO

            Raises
            ------
            TODO

            Notes
            -----
            TODO: include mathematical description and explanatory image

            See Also
            --------
            abmodel.agent.execution_modes.ExecutionModes : TODO complete
            explanation

            check_field_existance : TODO complete explanation

            contagion_function : TODO complete explanation

            determine_disease_state_max_time : TODO complete explanation

            Examples
            --------
            TODO: include some examples
        """
        try:
            if execmode == ExecutionModes.iterative.value:
                df_copy = df.copy()

                df[["disease_state", "times_infected", "infected_by",
                    "disease_state_time", "do_calculate_max_time",
                    "do_update_immunization_params"]] = df.apply(
                    lambda row: contagion_function(
                        row["agent"],
                        row["x"],
                        row["y"],
                        row["immunization_level"],
                        row["key"],
                        row["disease_state"],
                        row["susceptibility_group"],
                        row["times_infected"],
                        row["disease_state_time"],
                        row["reduction_factor"],
                        natural_history,
                        disease_groups,
                        susceptibility_groups,
                        kdtree_by_disease_state,
                        agents_labels_by_disease_state,
                        df_copy
                        ),
                    axis=1
                    )
            elif execmode == ExecutionModes.dask.value:
                df_copy = df.copy()
                df = from_pandas(df, npartitions=npartitions)

                df[["disease_state", "times_infected", "infected_by",
                    "disease_state_time", "do_calculate_max_time",
                    "do_update_immunization_params"]] = df.apply(
                    lambda row: contagion_function(
                        row["agent"],
                        row["x"],
                        row["y"],
                        row["immunization_level"],
                        row["key"],
                        row["disease_state"],
                        row["susceptibility_group"],
                        row["times_infected"],
                        row["disease_state_time"],
                        row["reduction_factor"],
                        natural_history,
                        disease_groups,
                        susceptibility_groups,
                        kdtree_by_disease_state,
                        agents_labels_by_disease_state,
                        df_copy
                        ),
                    axis=1,
                    meta={
                        0: "str",
                        1: "int64",
                        2: "object",
                        3: "float64",
                        4: "bool",
                        5: "bool"
                    }
                    )
                df = df.compute()
            else:
                raise NotImplementedError(
                    f"`execmode = {execmode}` is still not implemented yet"
                    )

            # Call determine_disease_state_max_time function in order
            # to calculate it for those agents who went through a transition
            df = cls.determine_disease_state_max_time(
                df, disease_groups, natural_history, execmode
                )

            # Call update_immunization_params function in order to update them
            # for those agents who went through a transition
            df = cls.update_immunization_params(
                df, natural_history, execmode
                )

            # Update key column
            df = cls.generate_key_col(df, execmode)

        except Exception as error:
            validation_list = ["agent", "x", "y", "immunization_level",
                               "key", "disease_state", "susceptibility_group",
                               "times_infected", "disease_state_time",
                               "reduction_factor"]
            exception_burner([
                error,
                check_field_existance(df, validation_list)
                ])
        else:
            return df

    @classmethod
    def update_immunization_params(
        cls,
        df: DataFrame,
        natural_history: NaturalHistory,
        execmode: ExecutionModes = ExecutionModes.iterative.value,
        npartitions: Optional[int] = 1
    ) -> DataFrame:
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            Returns
            -------
            TODO

            Raises
            ------
            TODO

            Notes
            -----
            TODO: include mathematical description and explanatory image

            See Also
            --------
            abmodel.agent.execution_modes.ExecutionModes : TODO complete
            explanation

            check_field_existance : TODO complete explanation

            update_immunization_params_iterative : TODO complete explanation

            Examples
            --------
            TODO: include some examples
        """
        try:
            if execmode == ExecutionModes.iterative.value:
                df[["immunization_level", "immunization_slope",
                    "immunization_time", "immunization_max_time",
                    "do_update_immunization_params"]] = df.apply(
                    lambda row: update_immunization_params_iterative(
                        row["key"],  # In this case, this key is the old one
                        row["disease_state"],
                        row["immunization_level"],
                        row["immunization_slope"],
                        row["immunization_time"],  # In scale of days
                        row["immunization_max_time"],  # In scale of days
                        row["do_update_immunization_params"],
                        natural_history
                        ),
                    axis=1
                    )
            elif execmode == ExecutionModes.dask.value:
                df = from_pandas(df, npartitions=npartitions)
                df[["immunization_level", "immunization_slope",
                    "immunization_time", "immunization_max_time",
                    "do_update_immunization_params"]] = df.apply(
                    lambda row: update_immunization_params_iterative(
                        row["key"],  # In this case, this key is the old one
                        row["disease_state"],
                        row["immunization_level"],
                        row["immunization_slope"],
                        row["immunization_time"],  # In scale of days
                        row["immunization_max_time"],  # In scale of days
                        row["do_update_immunization_params"],
                        natural_history
                        ),
                    axis=1,
                    meta={
                        0: "float64",
                        1: "float64",
                        2: "float64",
                        3: "float64",
                        4: "bool"
                    }
                    )
                df = df.compute()
            else:
                raise NotImplementedError(
                    f"`execmode = {execmode}` is still not implemented yet"
                    )
        except Exception as error:
            validation_list = ["key", "disease_state", "immunization_level",
                               "immunization_slope", "immunization_time",
                               "immunization_max_time",
                               "do_update_immunization_params"]
            exception_burner([
                error,
                check_field_existance(df, validation_list)
                ])
        else:
            return df

    @classmethod
    def update_immunization_level(
        cls,
        df: DataFrame,
        dt: float,  # In scale of days
        natural_history: NaturalHistory,
        execmode: ExecutionModes = ExecutionModes.iterative.value,
        npartitions: Optional[int] = 1
    ) -> DataFrame:
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            Returns
            -------
            TODO

            Raises
            ------
            TODO

            Notes
            -----
            TODO: include mathematical description and explanatory image

            See Also
            --------
            abmodel.agent.execution_modes.ExecutionModes : TODO complete
            explanation

            check_field_existance : TODO complete explanation

            update_immunization_level_iterative : TODO complete explanation

            Examples
            --------
            TODO: include some examples
        """
        try:
            if execmode == ExecutionModes.iterative.value:
                df[["immunization_level", "immunization_slope",
                    "immunization_time", "immunization_max_time"]] = df.apply(
                    lambda row: update_immunization_level_iterative(
                        dt,  # In scale of days
                        row["immunization_level"],
                        row["immunization_slope"],
                        row["immunization_time"],  # In scale of days
                        row["immunization_max_time"]  # In scale of days
                        ),
                    axis=1
                    )
            elif execmode == ExecutionModes.dask.value:
                df = from_pandas(df, npartitions=npartitions)
                df[["immunization_level", "immunization_slope",
                    "immunization_time", "immunization_max_time"]] = df.apply(
                    lambda row: update_immunization_level_iterative(
                        dt,  # In scale of days
                        row["immunization_level"],
                        row["immunization_slope"],
                        row["immunization_time"],  # In scale of days
                        row["immunization_max_time"]  # In scale of days
                        ),
                    axis=1,
                    meta={
                        0: "float64",
                        1: "float64",
                        2: "float64",
                        3: "float64",
                    }
                    )
                df = df.compute()
            else:
                raise NotImplementedError(
                    f"`execmode = {execmode}` is still not implemented yet"
                    )
        except Exception as error:
            validation_list = ["key", "immunization_level",
                               "immunization_slope", "immunization_time",
                               "immunization_max_time"]
            exception_burner([
                error,
                check_field_existance(df, validation_list)
                ])
        else:
            return df

    @classmethod
    def update_alertness_state(
        cls,
        df: DataFrame,
        kdtree_by_disease_state: dict,
        agents_labels_by_disease_state: dict,
        natural_history: NaturalHistory,
        disease_groups: DiseaseStates,
        dead_disease_group: str,
        execmode: ExecutionModes = ExecutionModes.iterative.value,
        npartitions: Optional[int] = 1
    ) -> DataFrame:
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            Returns
            -------
            TODO

            Raises
            ------
            TODO

            Notes
            -----
            TODO: include mathematical description and explanatory image

            See Also
            --------
            abmodel.agent.execution_modes.ExecutionModes : TODO complete
            explanation

            check_field_existance : TODO complete explanation

            alertness_function : TODO complete explanation

            Examples
            --------
            TODO: include some examples
        """
        try:
            if execmode == ExecutionModes.iterative.value:
                df[["is_alert", "alerted_by"]] = df.apply(
                    lambda row: alertness_function(
                        row["agent"],
                        row["key"],
                        row["x"],
                        row["y"],
                        row["is_dead"],
                        row["vulnerability_group"],
                        row["disease_state"],
                        natural_history,
                        disease_groups,
                        kdtree_by_disease_state,
                        agents_labels_by_disease_state,
                        dead_disease_group
                        ),
                    axis=1
                    )
            elif execmode == ExecutionModes.dask.value:
                df = from_pandas(df, npartitions=npartitions)
                df[["is_alert", "alerted_by"]] = df.apply(
                    lambda row: alertness_function(
                        row["agent"],
                        row["key"],
                        row["x"],
                        row["y"],
                        row["is_dead"],
                        row["vulnerability_group"],
                        row["disease_state"],
                        natural_history,
                        disease_groups,
                        kdtree_by_disease_state,
                        agents_labels_by_disease_state,
                        dead_disease_group
                        ),
                    axis=1,
                    meta={
                        0: "bool",
                        1: "object"
                    }
                    )
                df = df.compute()
            else:
                raise NotImplementedError(
                    f"`execmode = {execmode}` is still not implemented yet"
                    )
        except Exception as error:
            validation_list = ["agent", "key", "x", "y", "is_dead",
                               "vulnerability_group", "disease_state"]
            exception_burner([
                error,
                check_field_existance(df, validation_list)
                ])
        else:
            return df

    @classmethod
    def apply_mobility_restrictions(
        cls,
        step: int,
        df: DataFrame,
        # mr_groups: MobilityGroups,
        beta: float,
        mrt_policies: Optional[dict] = None,  # MRTracingPolicies
        mrt_policies_df: Optional[DataFrame] = None,
        global_cyclic_mr: Optional[GlobalCyclicMR] = None,
        cyclic_mr_policies: Optional[dict] = None,  # CyclicMRPolicies
        cmr_policies_df: Optional[DataFrame] = None,
        grace_time_in_steps: Optional[int] = None,
        iteration_time: Optional[timedelta] = None,
        mr_adherence_groups: Optional[MRAdherenceGroups] = None,
        execmode: ExecutionModes = ExecutionModes.iterative.value,
        npartitions: Optional[int] = 1
    ) -> tuple[DataFrame, DataFrame, DataFrame]:
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            Returns
            -------
            TODO

            Raises
            ------
            TODO

            Notes
            -----
            TODO: include mathematical description and explanatory image

            See Also
            --------
            abmodel.agent.execution_modes.ExecutionModes : TODO complete
            explanation

            check_field_existance : TODO complete explanation

            _function : TODO complete explanation

            Examples
            --------
            TODO: include some examples
        """
        try:
            if execmode == ExecutionModes.iterative.value:
                # =============================================================
                # Mobility Restrictions - Tracing Policies
                # =============================================================

                # Init mrt_target_groups
                mrt_target_groups = []

                if mrt_policies is not None:
                    # Init policies_status dict
                    policies_status = {"step": step}

                    # Init target_groups_lists
                    target_groups_lists = []

                    for variable, policie in zip(
                        mrt_policies.keys(),
                        mrt_policies.values()
                    ):
                        former_activation_status = mrt_policies_df[
                            variable.value].iloc[-1]

                        if former_activation_status == "disabled":
                            # Find the corresponding count
                            # to assess start level
                            if variable == InterestVariables.dead:
                                count = df[["is_dead"]].sum()["is_dead"]
                            if variable == InterestVariables.diagnosed:
                                count = df[["is_diagnosed"]].sum()[
                                    "is_diagnosed"]
                            if variable == InterestVariables.ICU:
                                count = df[["is_in_ICU"]].sum()["is_in_ICU"]
                            if variable == InterestVariables.hospital:
                                count = df[["is_hospitalized"]].sum()[
                                    "is_hospitalized"]

                            if count >= policie.mr_start_level:
                                new_activation_status = "enabled"
                            else:
                                new_activation_status = "disabled"

                        if former_activation_status == "enabled":

                            mr_stop_mode = policie.mr_stop_mode

                            if mr_stop_mode == MRTStopModes.level_number:
                                # Find the corresponding count
                                # to assess stop level
                                if variable == InterestVariables.dead:
                                    count = df[["is_dead"]].sum()["is_dead"]
                                if variable == InterestVariables.diagnosed:
                                    count = df[["is_diagnosed"]].sum()[
                                        "is_diagnosed"]
                                if variable == InterestVariables.ICU:
                                    count = df[["is_in_ICU"]].sum()[
                                        "is_in_ICU"]
                                if variable == InterestVariables.hospital:
                                    count = df[["is_hospitalized"]].sum()[
                                        "is_hospitalized"]

                                if count <= policie.mr_stop_level:
                                    new_activation_status = "disabled"
                                else:
                                    new_activation_status = "enabled"

                            if mr_stop_mode == MRTStopModes.length:
                                # Find the step when the policie was enabled
                                df_copy = mrt_policies_df[
                                    ["step", variable.value]
                                    ].copy()

                                df_copy["status_changed"] = df_copy[
                                    variable.value].shift() != df_copy[
                                        variable.value]

                                enabled_start = df_copy[
                                    df_copy["status_changed"]
                                    ]["step"].iloc[-1]

                                # Here "step" is the current step
                                enabled_length = step - enabled_start

                                policie.set_mr_length_in_steps(
                                    iteration_time
                                )
                                mr_length = policie.mr_length_in_steps
                                if enabled_length >= mr_length:
                                    new_activation_status = "disabled"
                                else:
                                    # enabled_length < mr_length
                                    new_activation_status = "enabled"

                        # Append to policies_status dict
                        policies_status[variable.value] = \
                            [new_activation_status]

                        # Append target_groups_lists
                        if new_activation_status == "enabled":
                            target_groups_lists.append(policie.target_groups)

                    # Append policies_status to mrt_policies_df
                    mrt_policies_df = concat(
                        [mrt_policies_df, DataFrame(policies_status)],
                        ignore_index=True
                    )

                    # Flatten list of lists using list comprehension
                    flatten_target_groups_lists = [
                        item
                        for sublist in target_groups_lists
                        for item in sublist
                        ]

                    # Remove duplicates
                    mrt_target_groups = list(set(flatten_target_groups_lists))

                # =============================================================
                # Mobility Restrictions - Cyclic Policies
                # =============================================================

                # Init cmr_target_groups
                cmr_target_groups = []

                cond_1 = global_cyclic_mr is not None
                cond_2 = cyclic_mr_policies is not None

                if cond_1 and cond_2:
                    # Init policies_status dict
                    policies_status = {"step": step}

                    # Init target_groups_lists
                    target_groups_lists = []

                    global_cyclic_mr.set_global_mr_length(
                        iteration_time
                    )

                    if step > grace_time_in_steps:
                        global_activation_satus = \
                            cmr_policies_df["global_mr"].iloc[-1]

                        if global_activation_satus == "disabled":
                            # Verify the status of unrestricted_time_steps
                            if global_cyclic_mr.\
                                    unrestricted_time_steps is not None:
                                pass
                            else:
                                global_cyclic_mr.\
                                    set_unrestricted_time(
                                        iteration_time
                                    )
                            unrestricted_time_steps = \
                                global_cyclic_mr.unrestricted_time_steps

                            #  Calculate the elapsed time disabled
                            cmr_policies_df_copy = cmr_policies_df[
                                ["step", "global_mr"]].copy()

                            cmr_policies_df_copy["status_change"] = \
                                cmr_policies_df_copy["global_mr"].shift() != \
                                cmr_policies_df["global_mr"]

                            disabled_start_step = \
                                cmr_policies_df_copy[cmr_policies_df_copy[
                                    "status_change"]]["step"].iloc[-1]

                            disabled_steps = step - disabled_start_step
                            # Assign the policies_status for global_mr
                            # comparing with the elapsed time
                            policies_status["global_mr"] = \
                                "disabled" if disabled_steps < \
                                unrestricted_time_steps else "enabled"
                            # If enables, assign enabled_steps
                            if policies_status["global_mr"] == "enabled":
                                enabled_steps = 0

                            # In random mode, set unrestricted_time_steps
                            # to None and enabled_steps = 0 when there will be
                            # a change of policies_status to enabled
                            if global_cyclic_mr.unrestricted_time_mode == \
                                    CyclicMRModes.random:
                                if policies_status["global_mr"] == "enabled":
                                    global_cyclic_mr.\
                                        set_none_unrestricted_time()
                                    enabled_steps = 0
                                else:
                                    pass

                        if global_activation_satus == "enabled":
                            # Verify the status of unrestricted_time_steps
                            # Set None for random mode
                            global_mr_length_steps = \
                                global_cyclic_mr.global_mr_length_steps

                            # Calculate the elapsed time enabled
                            cmr_policies_df_copy = cmr_policies_df[
                                ["step", "global_mr"]].copy()

                            cmr_policies_df_copy["status_change"] = \
                                cmr_policies_df_copy["global_mr"].shift() != \
                                cmr_policies_df["global_mr"]

                            enabled_start_step = \
                                cmr_policies_df_copy[cmr_policies_df_copy[
                                    "status_change"]]["step"].iloc[-1]

                            enabled_steps = step - enabled_start_step

                            # Check if continue enabled or set disabled
                            policies_status["global_mr"] = \
                                "enabled" if enabled_steps < \
                                global_mr_length_steps else "disabled"

                            # In random mode, set unrestricted_time_steps
                            # when there will be a change of policies_status
                            # to disabled
                            if global_cyclic_mr.unrestricted_time_mode == \
                                    CyclicMRModes.random:
                                if enabled_steps == global_mr_length_steps:
                                    global_cyclic_mr.\
                                        set_unrestricted_time(
                                            iteration_time
                                        )
                        # Check the status of each group
                        if policies_status["global_mr"] == "enabled":

                            for group in cyclic_mr_policies.keys():
                                # Calling the methods so as to set mr_length
                                # and time_without_restrictions in steps
                                cyclic_mr_policies[group].\
                                    set_mr_length(iteration_time)

                                cyclic_mr_policies[group].\
                                    set_time_without_restrictions(
                                        iteration_time
                                    )

                                delay = cyclic_mr_policies[group].delay
                                mr_length = cyclic_mr_policies[group].\
                                    mr_length_in_steps

                                if delay:
                                    # Assign delay in steps
                                    cyclic_mr_policies[
                                        group
                                        ].set_delay(iteration_time)
                                    delay = cyclic_mr_policies[
                                            group
                                        ].delay_in_steps

                                    # Compare and append to policies_status
                                    # dict
                                    if delay > enabled_steps:
                                        policies_status[group] = "disabled"
                                    else:
                                        # Compare and append to policies_status
                                        # dict
                                        if delay < enabled_steps and \
                                                enabled_steps <= mr_length:
                                            policies_status[group] = "enabled"
                                        else:
                                            policies_status[group] = "disabled"
                                else:
                                    # Append to policies_status dict
                                    policies_status[group] = \
                                        "enabled" if enabled_steps <= \
                                        mr_length else "disabled"
                        else:
                            # Append to policies_status dict diabled for all
                            # groups
                            for col in setdiff1d(
                                cmr_policies_df.columns,
                                "step"
                            ):
                                policies_status[col] = "disabled"

                    elif step == grace_time_in_steps:
                        # Append to policies_status dict
                        policies_status["global_mr"] = "enabled"

                        for group in cyclic_mr_policies.keys():
                            if cyclic_mr_policies[group].delay > 0:
                                # Append to policies_status dict
                                policies_status[group] = "disabled"
                            else:
                                # Append to policies_status dict
                                policies_status[group] = "enabled"
                    else:
                        # Append to policies_status dict
                        for col in setdiff1d(cmr_policies_df.columns, "step"):
                            policies_status[col] = "disabled"

                    # Append policies_status to cmr_policies_df
                    cmr_policies_df = concat(
                        [cmr_policies_df, DataFrame(
                            policies_status, index=[0]
                            )],
                        ignore_index=True
                    )

                    policies_status_copy = policies_status.copy()
                    policies_status_copy.pop("step")
                    policies_status_copy.pop("global_mr")
                    policies_status_copy = zip(
                        policies_status_copy.keys(),
                        policies_status_copy.values()
                    )

                    # Append cmr_target_groups
                    for group, status in policies_status_copy:
                        if status == "enabled":
                            cmr_target_groups.append(group)

                # Create mrc_target_groups list
                mrc_target_groups = \
                    cmr_target_groups + mrt_target_groups

                # Remove duplicates
                mrc_target_groups = list(set(mrc_target_groups))

                df = cls.to_isolate_agents_by_mr(
                    df,
                    mrc_target_groups,
                    beta,
                    mr_adherence_groups,
                    execmode
                )
            else:
                raise NotImplementedError(
                    f"`execmode = {execmode}` is still not implemented yet"
                    )
        except Exception as error:
            validation_list = ["agent"]
            exception_burner([
                error,
                check_field_existance(df, validation_list)
                ])
        else:
            return df, mrt_policies_df, cmr_policies_df



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['abmodel',
 'abmodel.agent',
 'abmodel.analysis',
 'abmodel.models',
 'abmodel.population',
 'abmodel.utils',
 'abmodel.utils.helpers']

package_data = \
{'': ['*']}

install_requires = \
['dask>=2022.4.0,<2023.0.0',
 'munch>=2.5.0,<3.0.0',
 'numpy>=1.21,<2.0',
 'pandas>=1.2.4,<2.0.0',
 'pydantic>=1.8.2,<2.0.0',
 'scikit-learn>=0.24.2,<0.25.0',
 'scipy>=1.6.3,<2.0.0']

setup_kwargs = {
    'name': 'cdslib-agents',
    'version': '0.0.7',
    'description': 'Contagious diseases simulation using Agent-Based Models',
    'long_description': '# CDSLib - Agents\n\nWelcome to **CDSLib - Agents** package.\n\n[![Open Source Love svg2](https://badges.frapsoft.com/os/v2/open-source.svg?v=103)](https://github.com/ellerbrock/open-source-badges/)\n[![PyPI](https://img.shields.io/pypi/v/cdslib-agents?color=color=%2310d510)](https://pypi.org/project/cdslib-agents/)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/cdslib-agents?color=%2310d510)](https://pypi.org/project/cdslib-agents/)\n[![Maintenance](https://img.shields.io/badge/Maintained%3F-yes-green.svg)](https://github.com/fenfisdi/cdslib_agents/graphs/commit-activity)\n[![GitHub contributors](https://img.shields.io/github/contributors/fenfisdi/cdslib_agents)](https://github.com/fenfisdi/cdslib_agents/graphs/contributors)\n\n## License\n\n[![license](https://img.shields.io/github/license/fenfisdi/cdslib_agents)](./LICENSE)\n\n## Sponsors\n\n| [![Sponsored by](https://img.shields.io/badge/sponsored%20by-UdeA-yellow)](https://www.udea.edu.co/) | [![Sponsored by](https://img.shields.io/badge/sponsored%20by-minciencias-yellow)](https://minciencias.gov.co/) | [![Sponsored by](https://img.shields.io/badge/sponsored%20by-sena-yellow)](https://www.sena.edu.co/) |\n| :---: | :---: | :---: |\n| <img src="https://raw.githubusercontent.com/fenfisdi/cdslib_agents/2335693f162e3cca97f7bba8591db9b0076a3823/images/Escudo-UdeA.svg" alt="UdeA logo" height="34"> | <img src="https://raw.githubusercontent.com/fenfisdi/cdslib_agents/1755d1fc4e45e94fcdc4275709b93fdf4eabd5f2/images/Minciencias_Colombia.svg" alt="Minciencias logo" width="160" height="34"> | <img src="https://raw.githubusercontent.com/fenfisdi/cdslib_agents/2335693f162e3cca97f7bba8591db9b0076a3823/images/Sena_Colombia_logo.svg" alt="Sena logo" height="34"> |\n\n\n## Code quality metrics\n\n[![Quality Gate Status](https://sonarcloud.io/api/project_badges/measure?project=fenfisdi_cdslib_agents&metric=alert_status)](https://sonarcloud.io/summary/new_code?id=fenfisdi_cdslib_agents)\n[![Bugs](https://sonarcloud.io/api/project_badges/measure?project=fenfisdi_cdslib_agents&metric=bugs)](https://sonarcloud.io/summary/new_code?id=fenfisdi_cdslib_agents)\n[![Code Smells](https://sonarcloud.io/api/project_badges/measure?project=fenfisdi_cdslib_agents&metric=code_smells)](https://sonarcloud.io/summary/new_code?id=fenfisdi_cdslib_agents)\n[![Coverage](https://sonarcloud.io/api/project_badges/measure?project=fenfisdi_cdslib_agents&metric=coverage)](https://sonarcloud.io/summary/new_code?id=fenfisdi_cdslib_agents)\n[![Duplicated Lines (%)](https://sonarcloud.io/api/project_badges/measure?project=fenfisdi_cdslib_agents&metric=duplicated_lines_density)](https://sonarcloud.io/summary/new_code?id=fenfisdi_cdslib_agents)\n[![Lines of Code](https://sonarcloud.io/api/project_badges/measure?project=fenfisdi_cdslib_agents&metric=ncloc)](https://sonarcloud.io/summary/new_code?id=fenfisdi_cdslib_agents)\n[![Maintainability Rating](https://sonarcloud.io/api/project_badges/measure?project=fenfisdi_cdslib_agents&metric=sqale_rating)](https://sonarcloud.io/summary/new_code?id=fenfisdi_cdslib_agents)\n[![Reliability Rating](https://sonarcloud.io/api/project_badges/measure?project=fenfisdi_cdslib_agents&metric=reliability_rating)](https://sonarcloud.io/summary/new_code?id=fenfisdi_cdslib_agents)\n[![Security Rating](https://sonarcloud.io/api/project_badges/measure?project=fenfisdi_cdslib_agents&metric=security_rating)](https://sonarcloud.io/summary/new_code?id=fenfisdi_cdslib_agents)\n[![Technical Debt](https://sonarcloud.io/api/project_badges/measure?project=fenfisdi_cdslib_agents&metric=sqale_index)](https://sonarcloud.io/summary/new_code?id=fenfisdi_cdslib_agents)\n[![Vulnerabilities](https://sonarcloud.io/api/project_badges/measure?project=fenfisdi_cdslib_agents&metric=vulnerabilities)](https://sonarcloud.io/summary/new_code?id=fenfisdi_cdslib_agents)\n\n## Description\n\nAs part of the **Contagious Disease Simulation Library**, the\n**CDSLib - Agents** package is intended to be used for modelling and simulating\ncontagious diseases using Agent-Based Models and it has been used for the development\nof the [Contagious Disease Simulation Laboratory](https://github.com/fenfisdi/cdslab).\n\nThe package consists of different implementations of a heterogeneous population\nof agents following rules of behavior that determine their movement and the evolution\nof their infectious and clinical states.\n\nThe current implementation allows Agents to move in a bidemensional space following\na distribution of velocities based on population data. This can be achieved thanks to\nthe inclusion Distribution module which is a wrapper for different tools that provides\ndistribution-like function support.\n\nIn regards to the disease states, the current implementation also allows the user to\nadd as much states as preferred, and to create custom natural history of the disease\n(i.e. the design a custom graph for evolution of disease states).\n\nFor more information, please refer to the official documentation of the project.\n\n## Testing\n\nAll the tests were developed with the python tool pytest. To see them properly, type\nin the command line:\n\n`$ poetry run pytest <name_of_test_file>`\n\nin the same ubication of the test file.\n\nFlags are used with pytest for modifying the aspect of the output\nreport. Typing flags using pytest might be tedious, however, it is possible modify the `pytest.init` file for\n[change defaulting command line options](https://docs.pytest.org/en/6.2.x/example/simple.html#how-to-change-command-line-options-defaults). This allows to adopt as many flags\nas necesary. In our case, it was adopted the flags: `-s`, `--no-header`, `-rA`, `-vv`,\n`--cov`, `--cov-branch`.\n\nDuring test execution with pytest, any output sent to stdout and stderr is\n[captured](https://docs.pytest.org/en/6.2.x/capture.html). The `-s` flag  \ndisable all capturing and lets the user look at all the outputs, this is very\nimportant because the description of all tests was typed in the docstring of each method.\n\nThe `--no-header` flag disables the [initial header](https://docs.pytest.org/en/stable/changelog.html#id33).\n\nThe `-rA` flag shows extra test [summary info](https://docs.pytest.org/en/6.2.x/usage.html#detailed-summary-report) of all the tests.\n\nThe `-vv` flag indicates to pytest run in [verbosity](https://docs.pytest.org/en/latest/how-to/output.html) mode, this controls all the pytest outputs.\n\nThe flag `--cov` produces a [coverage](https://pytest-cov.readthedocs.io/en/latest/#welcome-to-pytest-cov-s-documentation) reports. It shows the percentage of code lines covered by the test, while `--cov-branch` shows the branches covered. A branch is\na decision taken by the code when it finds an if sentence or similar.\n\nFor more information about [pytest flags](https://docs.pytest.org/en/6.2.x/usage.html).\n\nInstead of running one test, it is possible to run all the tests at the same time. You must be inside of `tests` directory, and type in the command line:\n\n`$ poetry run pytest`\n\nSo as to Modify the initial conditions of the test, use the fixtures methods at the beginning\nof all the tests files.\n\n![repo_logo](https://raw.githubusercontent.com/fenfisdi/cdslib_agents/main/images/CDSLib_agents_white-background.png "CDSLib - Agents Logo")\n\n## Authors and main contributors\n\n[![GitHub contributors](https://img.shields.io/github/contributors/fenfisdi/cdslib_agents)](https://github.com/fenfisdi/cdslib_agents/graphs/contributors)\n\nThis package is authored by \n[Camilo Hincapié](https://www.linkedin.com/in/camilo-hincapie-gutierrez/) (main author),\n[Ian Mejía](https://github.com/IanMejia),\n[Emil Rueda](https://www.linkedin.com/in/emil-rueda-424012207/),\n[Nicole Rivera](https://github.com/nicolerivera1)\nand\n[Carolina Rojas Duque](https://github.com/carolinarojasd)\nand the conceptual contributions about epidemiology of\n[Lina Marcela Ruiz Galvis](mailto:lina.ruiz2@udea.edu.co).\n\nOther remarkably contributors to this work were\n[Alejandro Campillo](https://www.linkedin.com/in/alucardcampillo/)\nand\n[Daniel Alfonso Montoya](https://www.linkedin.com/in/daniel-montoya-ds/).\n\n\n\n\n## Contact us\n\nFor any suggestion on the development of this type of models, please our official\nchannels of [dicussions](https://github.com/fenfisdi/cdslib_agents/discussions)\nprovided by GitHub.\n\n## Development\n\n### Create the virtualenv\n\nThis package is managed using [Pyenv](https://github.com/pyenv/pyenv) and \n[Poetry](https://python-poetry.org/docs/). In order to create the\nvirtualenv correctly use:\n\n```\n$ pyenv install 3.9.7\n$ poetry env use 3.9.7\n$ poetry install\n```\n\nThen activate the virtualenv running `poetry shell`, and to deactivate the virtual\nenvironment and exit this new shell type `exit`.\n',
    'author': 'Camilo Hincapié Gutiérrez',
    'author_email': 'camilo.hincapie@udea.edu.co',
    'maintainer': 'Vladimir Emil Rueda Gómez',
    'maintainer_email': 'vladimir.rueda@udea.edu.co',
    'url': 'https://pypi.org/project/cdslib-agents/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>3.9.7,<3.11',
}


setup(**setup_kwargs)



# Copyright (C) 2021, Camilo Hincapié Gutiérrez
# This file is part of CDSLIB.
#
# CDSLIB is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CDSLIB is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
#
# This package is authored by:
# Camilo Hincapié (https://www.linkedin.com/in/camilo-hincapie-gutierrez/) (main author)
# Ian Mejía (https://github.com/IanMejia)
# Emil Rueda (https://www.linkedin.com/in/emil-rueda-424012207/)
# Nicole Rivera (https://github.com/nicolerivera1)
# Carolina Rojas Duque (https://github.com/carolinarojasd)

from typing import Optional

from numpy import array, nan_to_num, inf, maximum, floor, setdiff1d, isin, pi
from scipy.spatial import KDTree
from pandas.core.frame import DataFrame
from pandas import concat

from abmodel.utils import Distribution
from abmodel.utils import ExecutionModes
from abmodel.utils import EvolutionModes
from abmodel.utils import timedelta_to_days
from abmodel.models import DistTitles
from abmodel.models import Configutarion
from abmodel.models import HealthSystem
from abmodel.models import SimpleGroups
from abmodel.models import SusceptibilityGroups
from abmodel.models import ImmunizationGroups
from abmodel.models import MobilityGroups
from abmodel.models import NaturalHistory
from abmodel.models import DiseaseStates
# from abmodel.models import MRTracingPolicies
from abmodel.models import GlobalCyclicMR
# from abmodel.models import CyclicMRPolicies
from abmodel.models import IsolationAdherenceGroups
from abmodel.models import MRAdherenceGroups
from abmodel.agent import AgentMovement
from abmodel.agent import AgentDisease
from abmodel.agent import AgentNeighbors
from .initial_arrangement import InitialArrangement


class Population:
    """
        TODO: Add brief explanation

        Attributes
        ----------
        TODO

        Methods
        -------
        TODO
    """
    def __init__(
        self,
        configuration: Configutarion,
        health_system: HealthSystem,
        age_groups: SimpleGroups,
        vulnerability_groups: SimpleGroups,
        mr_groups: SimpleGroups,
        susceptibility_groups: SusceptibilityGroups,
        mobility_groups: MobilityGroups,
        disease_groups: DiseaseStates,
        natural_history: NaturalHistory,
        initial_population_setup_list: list[dict],
        mrt_policies: Optional[dict] = None,  # MRTracingPolicies
        global_cyclic_mr: Optional[GlobalCyclicMR] = None,
        cyclic_mr_policies: Optional[dict] = None,  # CyclicMRPolicies
        immunization_groups: Optional[ImmunizationGroups] = None,
        isolation_adherence_groups: Optional[IsolationAdherenceGroups] = None,
        mr_adherence_groups: Optional[MRAdherenceGroups] = None,
        execmode: ExecutionModes = ExecutionModes.iterative.value,
        evolmode: EvolutionModes = EvolutionModes.steps.value
    ) -> None:
        """
            Constructor of Population class.

            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            See Also
            --------
            get_disease_groups_alive : TODO complete explanation

            choose_tracing_radius : TODO complete explanation

            Examples
            --------
            TODO: include some examples
        """
        # Store configuration
        self.configuration = configuration
        self.health_system = health_system
        self.age_groups = age_groups
        self.vulnerability_groups = vulnerability_groups
        self.mr_groups = mr_groups
        self.susceptibility_groups = susceptibility_groups
        self.mobility_groups = mobility_groups
        self.disease_groups = disease_groups
        self.natural_history = natural_history
        self.initial_population_setup_list = initial_population_setup_list
        self.mrt_policies = mrt_policies
        self.global_cyclic_mr = global_cyclic_mr
        self.cyclic_mr_policies = cyclic_mr_policies
        self.immunization_groups = immunization_groups
        self.isolation_adherence_groups = isolation_adherence_groups
        self.mr_adherence_groups = mr_adherence_groups
        self.execmode = execmode
        self.evolmode = evolmode

        # Required columns
        self.__req_cols_dict = {
            "age_group": self.age_groups,
            "disease_state": self.disease_groups,
            "mr_group": self.mr_groups,
            "vulnerability_group": self.vulnerability_groups,
            "mobility_group": self.mobility_groups,
            "susceptibility_group": self.susceptibility_groups,
            "immunization_group": self.immunization_groups,
            "isolation_adherence_group": self.isolation_adherence_groups,
            "mr_adherence_group": self.mr_adherence_groups
            }

        # TODO: Handle units

        # =====================================================================
        # Setup internal variables
        self.__get_disease_groups_alive()
        self.__choose_tracing_radius()

        # =====================================================================
        # Initialize population dataframe
        self.__initialize_df()

    @property
    def evolmode(self):
        return self.__evolmode

    @evolmode.setter
    def evolmode(self, value: EvolutionModes):
        self.__evolmode = value

    def __initialize_df(self):
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            See Also
            --------
            TODO
        """
        # =====================================================================
        # Init population dataframe __df
        self.__df = DataFrame({
            "agent": list(range(self.configuration.population_number))
        })

        # =====================================================================
        # Setup population's initial arrangement
        for initial_population_setup in self.initial_population_setup_list:
            self.__df = InitialArrangement.setup(
                self.__df,
                **initial_population_setup,
                )

        # Take care of required columns not initialized yet
        missing_cols = setdiff1d(
            list(self.__req_cols_dict.keys()),
            self.__df.columns.to_list()
            ).tolist()

        if missing_cols != []:
            self.__df = InitialArrangement.fulfill_setup(
                df=self.__df,
                missing_cols=missing_cols,
                req_cols_dict=self.__req_cols_dict
                )

        # =====================================================================
        # Store step information in a private attribute
        # for checking purposes
        self.__step = 0

        # Convert timedelta to days (float)
        self.dt = timedelta_to_days(self.configuration.iteration_time)

        # Initialize time columns
        self.__df.insert(loc=0, column="step", value=self.__step)
        self.__df.insert(loc=1, column="datetime",
                         value=self.configuration.initial_date)

        # =====================================================================
        # Initialize movement related columns
        self.__df = AgentMovement.init_required_fields(
            df=self.__df,
            box_size=self.configuration.box_size,
            mobility_groups=self.mobility_groups
            )

        # =====================================================================
        # Initialize disease related columns
        self.__df = AgentDisease.init_required_fields(
            df=self.__df,
            dead_disease_group=self.dead_disease_group,
            alpha=self.configuration.alpha,
            beta=self.configuration.beta,
            disease_groups=self.disease_groups,
            natural_history=self.natural_history,
            health_system=self.health_system,
            immunization_groups=self.immunization_groups,
            isolation_adherence_groups=self.isolation_adherence_groups,
            execmode=self.execmode
            )

        # =====================================================================
        # Initialize __accumulated_df
        if self.evolmode == EvolutionModes.cumulative.value:
            self.__accumulated_df = self.__df.copy()

        # =====================================================================
        # Initialize __mrt_policies_df
        if self.mrt_policies is not None:
            self.__mrt_policies_df = DataFrame({"step": [self.__step]})
            columns = [item.value for item in self.mrt_policies.keys()]
            for item in columns:
                self.__mrt_policies_df[item] = "disabled"

        else:
            self.__mrt_policies_df = None

        # =====================================================================
        # Initialize __cyclic_mr_policies_df
        cond_1 = self.global_cyclic_mr is not None
        cond_2 = self.cyclic_mr_policies is not None

        if cond_1 and cond_2:
            self.__grace_time_in_steps = int((
                self.global_cyclic_mr.grace_time
                - self.configuration.initial_date
                )/self.configuration.iteration_time)
            self.__cmr_policies_df = DataFrame({"step": [self.__step],
                "global_mr": ["disabled"]})
            for item in list(self.cyclic_mr_policies.keys()):
                self.__cmr_policies_df[item] = "disabled"
        else:
            self.__grace_time_in_steps = None
            self.__cmr_policies_df = None

    def get_population_df(self):
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            See Also
            --------
            TODO
        """
        return self.__df

    def get_accumulated_population_df(self):
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            See Also
            --------
            TODO

        """
        if self.evolmode == ExecutionModes.cumulative.value:
            return self.__accumulated_df
        else:
            raise ValueError(f"Denied: evolmode == {self.evolmode}")

    def get_mrt_policies_df(self):
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            See Also
            --------
            TODO
        """
        if self.mrt_policies is not None:
            return self.__mrt_policies_df
        else:
            raise ValueError("Denied: mrt_policies is None")

    def get_cmr_policies_df(self):
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            See Also
            --------
            TODO
        """
        cond_1 = self.global_cyclic_mr is not None
        cond_2 = self.cyclic_mr_policies is not None

        if cond_1 and cond_2:
            return self.__cmr_policies_df, self.__grace_time_in_steps
        else:
            if not cond_1 and not cond_2:
                raise ValueError(
                    "Denied: both global_cyclic_mr and "
                    "cyclic_mr_policies are None"
                    )
            if not cond_1:
                raise ValueError("Denied: global_cyclic_mr is None")
            if not cond_2:
                raise ValueError("Denied: cyclic_mr_policies is None")

    def get_units(self):
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            See Also
            --------
            TODO

            Examples
            --------
            TODO: include some examples
        """
        return None

    def evolve(
        self,
        iterations: int
    ):
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            See Also
            --------
            TODO

            Examples
            --------
            TODO: include some examples
        """
        # TODO: add the option to iterate by date
        for step in range(iterations):
            self.__evolve_single_step()

            if self.evolmode == EvolutionModes.cumulative.value:
                self.__accumulated_df = concat(
                    [self.__accumulated_df, self.__df],
                    ignore_index=True
                    )

    def __remove_dead_agents(self):
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            See Also
            --------
            TODO

            Examples
            --------
            TODO: include some examples
        """
        self.__df = self.__df[
            ~self.__df[["is_dead"]].any(axis="columns")
            ]

    def __evolve_single_step(self):
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            See Also
            --------
            TODO

            Examples
            --------
            TODO: include some examples
        """
        # =====================================================================
        # Remove dead agents before evolving population dataframe
        self.__remove_dead_agents()
        former_df = self.__df.copy()

        # =====================================================================
        # Evolve step
        self.__step += 1
        self.__df["step"] = self.__step

        # Evolve date
        self.__df["datetime"] += self.configuration.iteration_time

        # =====================================================================
        # Update agents' velocities
        for mobility_group in self.mobility_groups.items.keys():
            self.__df = AgentMovement.update_velocities(
                df=self.__df,
                distribution=self.mobility_groups
                .items[mobility_group].dist[DistTitles.mobility.value],
                angle_variance=self.mobility_groups
                .items[mobility_group].angle_variance,
                group_field="mobility_group",
                group_label=mobility_group,
                preserve_dtypes_dict={"step": int, "agent": int}
                )

        # =====================================================================
        # Update population states by means of state transition
        self.__df = AgentDisease.disease_state_transition(
            df=self.__df,
            dt=self.dt,
            disease_groups=self.disease_groups,
            natural_history=self.natural_history,
            execmode=self.execmode
            )

        # =====================================================================
        # Update Hospitalization and ICU status
        self.__df = AgentDisease.to_hospitalize_agents(
            df=self.__df,
            dead_disease_group=self.dead_disease_group,
            alpha=self.configuration.alpha,
            disease_groups=self.disease_groups,
            health_system=self.health_system,
            execmode=ExecutionModes.vectorized.value
            )

        # =====================================================================
        # Update diagnosis status
        self.__df = AgentDisease.to_diagnose_agents(
            df=self.__df,
            disease_groups=self.disease_groups,
            execmode=self.execmode
            )

        # =====================================================================
        # Update isolation status
        self.__df = AgentDisease.to_isolate_agents(
            df=self.__df,
            dt=self.dt,
            beta=self.configuration.beta,
            disease_groups=self.disease_groups,
            isolation_adherence_groups=self.isolation_adherence_groups,
            execmode=self.execmode
            )

        # =====================================================================
        # Stop isolated and hospitalized agents
        indexes = self.__df.query(
            "is_isolated == True | is_hospitalized == True"
            ).index.values

        if len(indexes) != 0:
            self.__df = AgentMovement.stop_agents(self.__df, indexes)

        # =====================================================================
        # Initialize velocities for formerly isolated and formerly hospitalized
        # agents
        former_indexes = former_df.query(
            "is_isolated == True | is_hospitalized == True"
            ).index.values

        mask = isin(former_indexes, indexes, invert=True)
        should_init_indexes = former_indexes[mask]

        if len(should_init_indexes) != 0:
            for mobility_group in self.mobility_groups.items.keys():
                self.__df = AgentMovement.initialize_velocities(
                    df=self.__df,
                    distribution=self.mobility_groups.items[
                        mobility_group].dist[DistTitles.mobility.value],
                    angle_distribution=Distribution(
                        dist_type="numpy",
                        dist_name="uniform",
                        low=0.0,
                        high=2*pi
                        ),
                    indexes=should_init_indexes,
                    group_field="mobility_group",
                    group_label=mobility_group,
                    preserve_dtypes_dict={"step": int, "agent": int}
                    )

        # =====================================================================
        # Create KDTree for agents of each alive disease state
        self.__kdtrees_and_agents_indices()

        # =====================================================================
        # Trace neighbors to susceptible agents
        self.__df = AgentNeighbors.trace_neighbors_to_susceptibles(
            df=self.__df,
            tracing_radius=self.tracing_radius,
            kdtree_by_disease_state=self.kdtree_by_disease_state,
            agents_labels_by_disease_state=self.agents_labels_by_disease_state,
            dead_disease_group=self.dead_disease_group,
            disease_groups=self.disease_groups,
            execmode=ExecutionModes.vectorized.value
            )

        # =====================================================================
        # Update alertness states for avoiding avoidable agents
        self.__df = AgentDisease.update_alertness_state(
            df=self.__df,
            kdtree_by_disease_state=self.kdtree_by_disease_state,
            agents_labels_by_disease_state=self.agents_labels_by_disease_state,
            natural_history=self.natural_history,
            disease_groups=self.disease_groups,
            dead_disease_group=self.dead_disease_group,
            execmode=self.execmode
            )

        # =====================================================================
        # Change population states by means of contagion
        self.__df = AgentDisease.disease_state_transition_by_contagion(
            df=self.__df,
            kdtree_by_disease_state=self.kdtree_by_disease_state,
            agents_labels_by_disease_state=self.agents_labels_by_disease_state,
            natural_history=self.natural_history,
            disease_groups=self.disease_groups,
            susceptibility_groups=self.susceptibility_groups,
            execmode=self.execmode
            )

        # =====================================================================
        # Update immunization level
        self.__df = AgentDisease.update_immunization_level(
            df=self.__df,
            dt=self.dt,
            natural_history=self.natural_history,
            execmode=self.execmode
            )

        # =====================================================================
        # Mobility Restrictions
        variables = AgentDisease.apply_mobility_restrictions(
            step=self.__step,
            df=self.__df,
            beta=self.configuration.beta,
            mrt_policies=self.mrt_policies,
            mrt_policies_df=self.__mrt_policies_df,
            global_cyclic_mr=self.global_cyclic_mr,
            cyclic_mr_policies=self.cyclic_mr_policies,
            cmr_policies_df=self.__cmr_policies_df,
            grace_time_in_steps=self.__grace_time_in_steps,
            iteration_time=self.configuration.iteration_time,
            mr_adherence_groups=self.mr_adherence_groups,
            execmode=self.execmode
            )
        # Assign values
        self.__df = variables[0]
        self.__mrt_policies_df = variables[1]
        self.__cmr_policies_df = variables[2]
        
        # =====================================================================
        # Stop agents isolated by mobility resytictions
        indexes = self.__df.query("isolated_by_mr == True").index.values

        if len(indexes) != 0:
            self.__df = AgentMovement.stop_agents(self.__df, indexes)

        # =====================================================================
        # Initialize velocities for formerly isolated by mr
        # agents
        former_indexes = former_df.query(
            "isolated_by_mr == True"
            ).index.values

        mask = isin(former_indexes, indexes, invert=True)
        should_init_indexes = former_indexes[mask]

        if len(should_init_indexes) != 0:
            for mobility_group in self.mobility_groups.items.keys():
                self.__df = AgentMovement.initialize_velocities(
                    df=self.__df,
                    distribution=self.mobility_groups.items[
                        mobility_group].dist[DistTitles.mobility.value],
                    angle_distribution=Distribution(
                        dist_type="numpy",
                        dist_name="uniform",
                        low=0.0,
                        high=2*pi
                        ),
                    indexes=should_init_indexes,
                    group_field="mobility_group",
                    group_label=mobility_group,
                    preserve_dtypes_dict={"step": int, "agent": int}
                    )

        # =====================================================================
        # Avoid avoidable agents
        filtered_df = self.__df[self.__df["is_alert"]][["agent", "alerted_by"]]
        df_to_avoid = filtered_df \
            .explode("alerted_by") \
            .rename(columns={"alerted_by": "agent_to_avoid"})

        self.__df = AgentMovement.avoid_agents(
            df=self.__df,
            df_to_avoid=df_to_avoid
            )

        # =====================================================================
        # Update agents' positions
        # NOTE: mobility_profile should be in iteration_time units, so the
        # value we should use for dt here is dt=1.0
        self.__df = AgentMovement.move_agents(
            df=self.__df,
            box_size=self.configuration.box_size,
            dt=1.0
            )

    def __get_disease_groups_alive(self) -> None:
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            See Also
            --------
            TODO

            Examples
            --------
            TODO: include some examples
        """
        # Retrieve disease group label which corresponds to those dead
        for disease_group_label in self.disease_groups.items.keys():
            if self.disease_groups.items[disease_group_label].is_dead:
                self.dead_disease_group = disease_group_label

        # Also exclude those dead from disease groups alive
        self.disease_groups_alive = list(setdiff1d(
            list(self.disease_groups.items.keys()),
            [self.dead_disease_group]
            ))

    def __choose_tracing_radius(self) -> None:
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            See Also
            --------
            TODO

            Examples
            --------
            TODO: include some examples
        """
        # Retrieve maximum radius for trace_neighbors function
        spread_radius_list = [
            self.disease_groups.items[disease_group].spread_radius
            for disease_group in self.disease_groups.items.keys()
            ]

        avoidance_radius_list = [
            self.natural_history.items[key].avoidance_radius
            for key in self.natural_history.items.keys()
            ]

        spread_radius_arr = array(spread_radius_list, dtype=float)
        spread_radius_arr = nan_to_num(spread_radius_arr, nan=-inf)
        max_spread_radius = spread_radius_arr.max()

        avoidance_radius_arr = array(avoidance_radius_list, dtype=float)
        avoidance_radius_arr = nan_to_num(avoidance_radius_arr, nan=-inf)
        max_avoidance_radius = avoidance_radius_arr.max()

        self.tracing_radius = maximum(
            max_spread_radius,
            max_avoidance_radius
            )

    def __kdtrees_and_agents_indices(self) -> None:
        """
            TODO: Add brief explanation

            Parameters
            ----------
            TODO

            See Also
            --------
            TODO

            Examples
            --------
            TODO: include some examples
        """
        self.kdtree_by_disease_state = {}
        self.agents_labels_by_disease_state = {}

        for disease_state in self.disease_groups_alive:
            # Filter population
            # Exclude those agents hospitalized and those that are dead
            filtered_df = self.__df.loc[
                (self.__df["disease_state"] == disease_state)
                &
                (~self.__df["is_hospitalized"])
                &
                (~self.__df["is_dead"])
                ][["agent", "x", "y"]].copy()

            # Calculate how many points were retrieved
            n_points = filtered_df.shape[0]

            if n_points != 0:
                # Get all agents locations
                locations = filtered_df[["x", "y"]].to_numpy()

                # Get all agents labels
                agents_labels = filtered_df["agent"].to_numpy()

                # Select a sensible leafsize for the KDtree method
                one_percent_of_points = floor(n_points*0.01)
                leafsize = (
                    one_percent_of_points
                    if one_percent_of_points > 10 else 10
                    )

                # Initialize (calculate) tree for the disease state
                # and store it inside the dict kdtree_by_disease_state
                self.kdtree_by_disease_state[disease_state] = KDTree(
                    locations, leafsize=leafsize
                    )

                # Also store the corresponding agents labels
                self.agents_labels_by_disease_state[disease_state] = \
                    agents_labels
            else:
                # n_points == 0
                self.kdtree_by_disease_state[disease_state] = None
                self.agents_labels_by_disease_state[disease_state] = None



# Copyright (C) 2021, Camilo Hincapié Gutiérrez
# This file is part of CDSLIB.
#
# CDSLIB is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CDSLIB is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
#
# This package is authored by:
# Camilo Hincapié (https://www.linkedin.com/in/camilo-hincapie-gutierrez/) (main author)
# Ian Mejía (https://github.com/IanMejia)
# Emil Rueda (https://www.linkedin.com/in/emil-rueda-424012207/)
# Nicole Rivera (https://github.com/nicolerivera1)
# Carolina Rojas Duque (https://github.com/carolinarojasd)

from time import time
from typing import Union, Any, Optional

from pydantic import validate_arguments
from numpy import abs, genfromtxt, ndarray, random, ones, full
from sklearn.neighbors import KernelDensity


class Distribution:
    """
        Distribution class

        It computes random numbers from a probability density distribution.

        Methods
        -------
        TODO
    """
    @validate_arguments(config={"arbitrary_types_allowed": True})
    def __init__(self,
                 dist_type: Any,
                 constant: Optional[float] = None,
                 data: Optional[ndarray] = None,
                 filename: Optional[str] = None,
                 dist_name: Optional[str] = None,
                 **kwargs):
        """
            Constructor of Distribution class.

            It specifies which type of distribution is going to be used
            and its corresponding parameters.

            Parameters
            ----------
            dist_type : {None, 'constant', 'empirical', 'weights', 'numpy'}

                None : it allows a `Distribution` object to be set in such
                        way that it always return None.

                'constant' : it numerically implements a "Dirac delta"
                    function, i.e. all points will have the same value
                    specified by the parameter `constant`

                'empirical' : build distributions from empirical data,
                    estimating the overall shape of the distribution using
                    the KDE approach available via Scikit-Learn. If the data is
                    stored in a file, then 'filename' must be passed to specify
                    the path to the file and the data inside that file must be
                    formatted without header in this way:

                    .. code-block::
                        data_0
                        data_1
                        data_2
                        ...

                'weights' : it needs the histogram (xi, pi) data
                    in order to generate a random sample from
                    this given array of points and their corresponding
                    probability weights.
                    It uses Numpy Random Choices [2]_. If the data is
                    stored in a file, then 'filename' must be passed to specify
                    the path to the file and the data inside that file must be
                    formatted without header in this way:

                    .. code-block::
                        x_0, p_0
                        x_1, p_1
                        x_2, p_2
                        ...

                'numpy' : it uses the distributions implemented in
                    Numpy Random Distributions [3]_

            constant : float, optional
                It specifies the constant value to which all point are
                mapped.

            data : ndarray, optional
                It corresponds

            filename : str, optional
                It specifies the path for the required data in order to build
                a distribution from this data. It is required by
                `dist_type = 'empirical'` or `dist_type = 'weights'`.

            dist_name : str, optional
                It specifies the distribution to use from numpy when
                `dist_type = 'numpy'`.

            **kwargs : dict, optional
                Extra arguments that must to be passed to the
                Scikit-Learn's Kernel Density constructor [1]_ or
                to the Numpy Random Distributions [3]_, except the parameter
                `size`. In the latter case, keyword argument must not be
                passed since it is used directly in the sampling methods.

            Raises
            ------
            ValueError
                TODO: when ?

            SystemError
                TODO: when ?

            References
            ----------
            .. [1] [Scikit-learn: Kernel Density Estimation](https://scikit-learn.org/stable/modules/density.html#kernel-density)
            .. [2] [Numpy Random Choice](https://numpy.org/doc/stable/reference/random/generated/numpy.random.Generator.choice.html#numpy.random.Generator.choice)
            .. [3] [Numpy Random Distributions](https://numpy.org/doc/stable/reference/random/generator.html#distributions)

            Examples
            --------
            TODO: include some exhaustive examples here using each dist_type
        """
        self.dist_type = dist_type
        self.seed = int(time())

        try:
            if self.dist_type is None:
                # Allows a `Distribution` object set to None
                self.constant = None

            elif self.dist_type == "constant":
                # "Dirac delta"-like function
                if constant is not None:
                    self.constant = constant
                else:
                    raise ValueError(
                        "Parameter `constant` should not be None when "
                        "`dist_type` = 'constant'"
                        )

            elif self.dist_type == "empirical":
                # Using KernelDensity estimator from Scikit-learn
                if data is not None and filename is None:
                    pass
                elif data is None and filename is not None:
                    self.filename = filename

                    data = genfromtxt(self.filename)
                else:
                    raise ValueError(
                        "The data is required and must be provided through "
                        "the parameter `data` or `filename` when "
                        "`dist_type` = 'empirical', then one of them should "
                        "not be None"
                        )

                # Check data has the right dimension
                if data.ndim == 1:
                    self.kwargs = kwargs
                    self.kd_estimator = KernelDensity(**self.kwargs).fit(
                        data.reshape(-1, 1)
                        )
                else:
                    raise ValueError(
                        "The data provided should be a 1-D array."
                        )

            elif self.dist_type == "weights":
                # Using numpy.random.choice
                if data is not None and filename is None:
                    pass
                elif data is None and filename is not None:
                    self.filename = filename
                    data = genfromtxt(self.filename, delimiter=",")
                else:
                    raise ValueError(
                        "The data is required and must be provided through "
                        "the parameter `data` or `filename` when "
                        "`dist_type` = 'weights', then one of them should "
                        "not be None"
                        )

                # Check data has the right dimension
                if data.ndim == 2:
                    self.xi = data[:, 0]
                    self.pi = data[:, 1]

                    self.random_number_generator = \
                        random.default_rng(seed=self.seed)
                else:
                    raise ValueError(
                        "The data provided should be a 2-D array."
                        )

            elif self.dist_type == "numpy":
                # Using numpy.random
                self.kwargs = kwargs

                self.random_number_generator = \
                    random.default_rng(seed=self.seed)

                if dist_name in dir(random):
                    self.numpy_distribution = getattr(
                        self.random_number_generator,
                        dist_name
                        )
                else:
                    raise ValueError(
                        f"Distribution '{dist_name}' is not "
                        "implemented in  numpy.random. See: "
                        "https://numpy.org/doc/stable/reference/random/"
                        "generator.html#distributions"
                        )
            else:
                raise ValueError(
                    "'dist_type' is not in "
                    "{None, 'constant', 'empirical', 'weights', 'numpy'}"
                    )

        except Exception as error:
            self.manage_exception(
                exception=error,
                message="Error initializing distribution."
                )

    @validate_arguments
    def sample(self, size: int = 1) -> Union[ndarray, int, float]:
        """
            Compute random sampling using the defined distribution.

            Parameters
            ----------
            size : int, default=1
                Number of samples to generate.

            Returns
            -------
            samples : int or numpy.array
                Samples generated.

            Raises
            ------
            SystemError
                TODO: when ?

            Examples
            --------
            TODO: include some examples
        """
        if self.dist_type is None:
            # Always return None
            samples = full(size, self.constant)

        elif self.dist_type == "constant":
            # "Dirac delta"-like function
            samples = self.constant*ones(size)
        elif self.dist_type == "empirical":
            # Using KernelDensity estimator from Scikit-learn
            try:
                samples = self.kd_estimator.sample(
                    n_samples=size, random_state=self.seed).flatten()
            except Exception as error:
                self.manage_exception(
                    exception=error,
                    message="Error using KernelDensity estimator."
                    )
        elif self.dist_type == "weights":
            # Using numpy.random.choice
            try:
                samples = self.random_number_generator.choice(
                    a=self.xi, p=self.pi, size=size)
            except Exception as error:
                self.manage_exception(
                    exception=error,
                    message="Exception with numpy random choice."
                    )

        elif self.dist_type == "numpy":
            # Using numpy.random
            try:
                samples = self.numpy_distribution(**self.kwargs, size=size)
            except Exception as error:
                self.manage_exception(
                    exception=error,
                    message="Exception with numpy random distributions."
                    )

        else:
            raise ValueError(
                "'dist_type' is not in "
                "{'constant', 'empirical', 'weights', 'numpy'}"
                )

        if size == 1:
            return samples[0]
        else:
            return samples

    @validate_arguments
    def sample_positive(self, size: int = 1) -> Union[ndarray, int]:
        """
            Compute one random sample using the defined distribution,
            but restrincting output to be strictly positive.

            Parameters
            ----------
            size : int, default=1

            Raises
            ------
            SystemError
                TODO: when ?

            See Also
            --------
            sample : Compute random sampling using the defined distribution.

            Examples
            --------
            TODO: include some examples
        """
        return abs(self.sample(size))

    @validate_arguments(config={"arbitrary_types_allowed": True})
    def manage_exception(self, exception: Exception, message: str):
        """
            Internal function for managing exceptions

            Parameters
            ----------
            exception : Exception
                Exception raised by the workflow.

            message : str
                String with the error message.

            Raises
            ------
            SystemError
                TODO: when ?

            Examples
            --------
            TODO: include some examples
        """
        raise SystemError(f"{message}\nError: {exception}")



# Copyright (C) 2021, Camilo Hincapié Gutiérrez
# This file is part of CDSLIB.
#
# CDSLIB is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# CDSLIB is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
#
# This package is authored by:
# Camilo Hincapié (https://www.linkedin.com/in/camilo-hincapie-gutierrez/) (main author)
# Ian Mejía (https://github.com/IanMejia)
# Emil Rueda (https://www.linkedin.com/in/emil-rueda-424012207/)
# Nicole Rivera (https://github.com/nicolerivera1)
# Carolina Rojas Duque (https://github.com/carolinarojasd)

from pydantic import BaseModel


class HealthSystem(BaseModel):
    """
        TODO: Add brief explanation

        Attributes
        ----------
        TODO

        Examples
        --------
        TODO: include some examples
    """
    hospital_capacity: int
    ICU_capacity: int
